# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import netCDF4
import pytest
from torch_geometric.data import HeteroData

from anemoi.graphs.edges import ICONTopologicalDecoderEdges
from anemoi.graphs.edges import ICONTopologicalEncoderEdges
from anemoi.graphs.edges import ICONTopologicalProcessorEdges
from anemoi.graphs.nodes import ICONCellGridNodes
from anemoi.graphs.nodes import ICONMultiMeshNodes


class TestEdgeBuilderDependencies:
    """Test that ICON edge builders depend on the presence of ICON node builders."""

    @pytest.fixture()
    def icon_graph(self, monkeypatch, icon_dataset_mock) -> HeteroData:
        """Return a HeteroData object with ICON node builders."""
        monkeypatch.setattr(netCDF4, "Dataset", icon_dataset_mock)
        data_node_builder = ICONCellGridNodes(name="data", grid_filename="test.nc", max_level=1)
        hidden_node_builder = ICONMultiMeshNodes(name="hidden", grid_filename="test.nc", max_level=1)

        graph = HeteroData()
        graph = data_node_builder.update_graph(graph, {})
        graph = hidden_node_builder.update_graph(graph, {})

        return graph

    def test_encoder(self, icon_graph: HeteroData):
        """Test that the `ICONTopologicalEncoderEdges` depends on the presence of ICON node builders."""
        edge_builder = ICONTopologicalEncoderEdges(source_name="data", target_name="hidden")
        assert ("data", "to", "hidden") not in icon_graph.edge_types

        icon_graph = edge_builder.update_graph(icon_graph)

        assert ("data", "to", "hidden") in icon_graph.edge_types
        assert hasattr(icon_graph["data", "to", "hidden"], "edge_index")

    def test_wrong_encoder(self, icon_graph: HeteroData):
        """Test that the `ICONTopologicalEncoderEdges` depends on the presence of ICON node builders."""
        edge_builder = ICONTopologicalEncoderEdges(source_name="hidden", target_name="data")

        with pytest.raises(AssertionError):
            edge_builder.update_graph(icon_graph)

    def test_decoder(self, icon_graph: HeteroData):
        """Test that the `ICONTopologicalDecoderEdges` depends on the presence of ICON node builders."""
        edge_builder = ICONTopologicalDecoderEdges(source_name="hidden", target_name="data")
        assert ("hidden", "to", "data") not in icon_graph.edge_types

        icon_graph = edge_builder.update_graph(icon_graph)

        assert ("hidden", "to", "data") in icon_graph.edge_types
        assert hasattr(icon_graph["hidden", "to", "data"], "edge_index")

    def test_wrong_decoder(self, icon_graph: HeteroData):
        """Test that the `ICONTopologicalDecoderEdges` depends on the presence of ICON node builders."""
        edge_builder = ICONTopologicalDecoderEdges(source_name="data", target_name="hidden")

        with pytest.raises(AssertionError):
            edge_builder.update_graph(icon_graph)

    def test_processor(self, icon_graph: HeteroData):
        """Test that the `ICONTopologicalProcessorEdges` depends on the presence of ICON node builders."""
        edge_builder = ICONTopologicalProcessorEdges(source_name="hidden", target_name="hidden")
        assert ("hidden", "to", "hidden") not in icon_graph.edge_types

        icon_graph = edge_builder.update_graph(icon_graph)

        assert ("hidden", "to", "hidden") in icon_graph.edge_types
        assert hasattr(icon_graph["hidden", "to", "hidden"], "edge_index")

    def test_wrong_processor(self, icon_graph: HeteroData):
        """Test that the `ICONTopologicalProcessorEdges` depends on the presence of ICON node builders."""
        edge_builder = ICONTopologicalProcessorEdges(source_name="data", target_name="hidden")

        with pytest.raises(AssertionError):
            edge_builder.update_graph(icon_graph)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from pathlib import Path
from typing import Optional
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import plotly.graph_objects as go
import torch
from matplotlib.colors import rgb2hex
from torch_geometric.data import HeteroData

from anemoi.graphs.plotting.prepare import compute_isolated_nodes
from anemoi.graphs.plotting.prepare import compute_node_adjacencies
from anemoi.graphs.plotting.prepare import edge_list
from anemoi.graphs.plotting.prepare import node_list

annotations_style = {"text": "", "showarrow": False, "xref": "paper", "yref": "paper", "x": 0.005, "y": -0.002}
plotly_axis_config = {"showgrid": False, "zeroline": False, "showticklabels": False}

LOGGER = logging.getLogger(__name__)


def plot_interactive_subgraph_2d(
    graph: HeteroData,
    edges_to_plot: tuple[str, str, str],
    out_file: Optional[Union[str, Path]] = None,
) -> None:
    """Plots a bipartite graph (bi-graph).

    This methods plots the bipartite graph passed in an interactive window (using Ploty).

    Parameters
    ----------
    graph : dict
        The graph to plot.
    edges_to_plot : tuple[str, str]
        Names of the edges to plot.
    out_file : str | Path, optional
        Name of the file to save the plot. Default is None.
    """
    source_name, _, target_name = edges_to_plot
    edge_x, edge_y = edge_list(graph, source_nodes_name=source_name, target_nodes_name=target_name)
    assert source_name in graph.node_types, f"edges_to_plot ({source_name}) should be in the graph"
    assert target_name in graph.node_types, f"edges_to_plot ({target_name}) should be in the graph"
    lats_source_nodes, lons_source_nodes = node_list(graph, source_name)
    lats_target_nodes, lons_target_nodes = node_list(graph, target_name)

    # Compute node adjacencies
    node_adjacencies = compute_node_adjacencies(graph, source_name, target_name)
    node_text = [f"# of connections: {x}" for x in node_adjacencies]

    edge_trace = go.Scattergeo(
        lat=edge_x,
        lon=edge_y,
        line={"width": 0.5, "color": "#888"},
        hoverinfo="none",
        mode="lines",
        name="Connections",
    )

    source_node_trace = go.Scattergeo(
        lat=lats_source_nodes,
        lon=lons_source_nodes,
        mode="markers",
        hoverinfo="text",
        name=source_name,
        marker={
            "showscale": False,
            "color": "red",
            "size": 2,
            "line_width": 2,
        },
    )

    target_node_trace = go.Scattergeo(
        lat=lats_target_nodes,
        lon=lons_target_nodes,
        mode="markers",
        hoverinfo="text",
        name=target_name,
        text=node_text,
        marker={
            "showscale": True,
            "colorscale": "YlGnBu",
            "reversescale": True,
            "color": list(node_adjacencies),
            "size": 10,
            "colorbar": {"thickness": 15, "title": "Node Connections", "xanchor": "left", "title_side": "right"},
            "line_width": 2,
        },
    )
    layout = go.Layout(
        title="<br>" + f"Graph {source_name} --> {target_name}",
        title_font_size=16,
        showlegend=True,
        hovermode="closest",
        margin={"b": 20, "l": 5, "r": 5, "t": 40},
        annotations=[annotations_style],
        legend={"x": 0, "y": 1},
        xaxis=plotly_axis_config,
        yaxis=plotly_axis_config,
    )
    fig = go.Figure(data=[edge_trace, source_node_trace, target_node_trace], layout=layout)
    fig.update_geos(fitbounds="locations")

    if out_file is not None:
        fig.write_html(out_file)
    else:
        fig.show()


def plot_isolated_nodes_2d(graph: HeteroData, out_file: Optional[Union[str, Path]] = None) -> None:
    """Plot isolated nodes.

    This method creates an interactive visualization of the isolated nodes in the graph.

    Parameters
    ----------
    graph : AnemoiGraph
        The graph to plot.
    out_file : str | Path, optional
        Name of the file to save the plot. Default is None.
    """
    isolated_nodes = compute_isolated_nodes(graph)

    if len(isolated_nodes) == 0:
        LOGGER.warning("No isolated nodes found.")
        return

    colorbar = plt.cm.rainbow(np.linspace(0, 1, len(isolated_nodes)))
    nodes = []
    for name, (lat, lon) in isolated_nodes.items():
        nodes.append(
            go.Scattergeo(
                lat=lat,
                lon=lon,
                mode="markers",
                hoverinfo="text",
                name=name,
                marker={"showscale": False, "color": rgb2hex(colorbar[len(nodes)]), "size": 10},
            ),
        )

    layout = go.Layout(
        title="<br>Orphan nodes",
        title_font_size=16,
        showlegend=True,
        hovermode="closest",
        margin={"b": 20, "l": 5, "r": 5, "t": 40},
        annotations=[annotations_style],
        legend={"x": 0, "y": 1},
        xaxis=plotly_axis_config,
        yaxis=plotly_axis_config,
    )
    fig = go.Figure(data=nodes, layout=layout)
    fig.update_geos(fitbounds="locations")

    if out_file is not None:
        fig.write_html(out_file)
    else:
        fig.show()


def plot_interactive_nodes_2d(graph: HeteroData, nodes_name: str, out_file: Optional[str] = None) -> None:
    """Plot nodes.

    This method creates an interactive visualization of a set of nodes.

    Parameters
    ----------
    graph : HeteroData
        Graph.
    nodes_name : str
        Name of the nodes to plot.
    out_file : str, optional
        Name of the file to save the plot. Default is None.
    """
    node_latitudes, node_longitudes = node_list(graph, nodes_name)
    node_attrs = graph[nodes_name].node_attrs()
    # Remove x to avoid plotting the coordinates as an attribute
    node_attrs.remove("x")

    if len(node_attrs) == 0:
        LOGGER.warning(f"No node attributes found for {nodes_name} nodes.")
        return

    node_traces = {}
    for node_attr in node_attrs:
        node_attr_values = graph[nodes_name][node_attr].float().numpy()

        if node_attr_values.ndim == 1:
            node_attr_values = torch.unsqueeze(node_attr_values, -1)

        for attr_dim in range(node_attr_values.shape[1]):
            suffix = "" if node_attr_values.shape[1] == 1 else f"_[{attr_dim}]"
            node_traces[node_attr + suffix] = go.Scattergeo(
                lat=node_latitudes,
                lon=node_longitudes,
                name=" ".join((node_attr + suffix).split("_")).capitalize(),
                mode="markers",
                hoverinfo="text",
                marker={
                    "color": node_attr_values[:, attr_dim].squeeze().tolist(),
                    "showscale": True,
                    "colorscale": "RdBu",
                    "colorbar": {"thickness": 15, "title": node_attr + suffix, "xanchor": "left"},
                    "size": 5,
                },
                visible=False,
            )

    # Create and add slider
    slider_steps = []
    for i, node_attr in enumerate(node_traces.keys()):
        step = dict(
            label=f"Node attribute: {node_attr}",
            method="update",
            args=[{"visible": [False] * len(node_traces)}],
        )
        step["args"][0]["visible"][i] = True  # Toggle i'th trace to "visible"
        slider_steps.append(step)

    fig = go.Figure(
        data=list(node_traces.values()),
        layout=go.Layout(
            title=f"<br>Map of {nodes_name} nodes",
            sliders=[
                dict(active=0, currentvalue={"visible": False}, len=0.4, x=0.5, xanchor="center", steps=slider_steps)
            ],
            title_font_size=16,
            showlegend=False,
            hovermode="closest",
            margin={"b": 20, "l": 5, "r": 5, "t": 40},
            annotations=[annotations_style],
            xaxis=plotly_axis_config,
            yaxis=plotly_axis_config,
        ),
    )
    fig.data[0].visible = True

    if out_file is not None:
        fig.write_html(out_file)
    else:
        fig.show()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

import torch
from torch_geometric.data import HeteroData
from torch_geometric.data.storage import NodeStorage

from anemoi.graphs.edges.builders.base import BaseEdgeBuilder
from anemoi.graphs.generate.icon_mesh import ICONCellDataGrid
from anemoi.graphs.generate.icon_mesh import ICONMultiMesh

LOGGER = logging.getLogger(__name__)


class ICONTopologicalProcessorEdges(BaseEdgeBuilder):
    """ICON Topological Processor Edges

    Computes edges based on ICON grid topology: processor grid built
    from ICON grid vertices.
    """

    def compute_edge_index(self, source_nodes: NodeStorage, target_nodes: NodeStorage) -> torch.Tensor:
        """Compute the edge indices for the KNN method.

        Parameters
        ----------
        source_nodes : NodeStorage
            The source nodes.
        target_nodes : NodeStorage
            The target nodes.

        Returns
        -------
        torch.Tensor of shape (2, num_edges)
            Indices of source and target nodes connected by an edge.
        """
        assert isinstance(
            source_nodes["_icon_nodes"], ICONMultiMesh
        ), f"{self.__class__.__name__}: source nodes must be ICONMultiMesh"
        assert isinstance(
            target_nodes["_icon_nodes"], ICONMultiMesh
        ), f"{self.__class__.__name__}: target nodes must be ICONMultiMesh"
        edge_index = source_nodes["_icon_nodes"].multi_mesh_edges
        return torch.from_numpy(edge_index.T)


class BaseICONEdgeBuilder(BaseEdgeBuilder):
    """Base ICON Edge Builder."""

    def prepare_node_data(self, graph: HeteroData) -> tuple[ICONCellDataGrid, ICONMultiMesh]:
        nodes_names = self.source_name, self.target_name
        cell_grid = graph[nodes_names[self.vertex_index[0]]]["_icon_nodes"]
        assert isinstance(
            cell_grid, ICONCellDataGrid
        ), f"{self.__class__.__name__}: source nodes must be ICONCellDataGrid"
        multi_mesh = graph[nodes_names[self.vertex_index[1]]]["_icon_nodes"]
        assert isinstance(multi_mesh, ICONMultiMesh), f"{self.__class__.__name__}: target nodes must be ICONMultiMesh"
        return cell_grid, multi_mesh

    def compute_edge_index(self, cell_grid: ICONCellDataGrid, multi_mesh: ICONMultiMesh) -> torch.Tensor:
        edge_vertices = cell_grid.get_grid2mesh_edges(multi_mesh)
        return torch.from_numpy(edge_vertices[:, self.vertex_index].T)


class ICONTopologicalEncoderEdges(BaseICONEdgeBuilder):
    """ICON Topological Encoder Edges

    Computes encoder edges based on ICON grid topology: ICON cell
    circumcenters for mapped onto processor grid built from ICON grid
    vertices.
    """

    vertex_index: tuple[int, int] = (0, 1)


class ICONTopologicalDecoderEdges(BaseICONEdgeBuilder):
    """ICON Topological Decoder Edges

    Computes decoder edges based on ICON grid topology: mapping from
    processor grid built from ICON grid vertices onto ICON cell
    circumcenters.
    """

    vertex_index: tuple[int, int] = (1, 0)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.nodes.attributes import SphericalAreaWeights
from anemoi.graphs.nodes.attributes import UniformWeights
from anemoi.graphs.nodes.builders.base import BaseNodeBuilder
from anemoi.graphs.nodes.builders.from_healpix import HEALPixNodes


@pytest.mark.parametrize("resolution", [2, 5, 7])
def test_init(resolution: int):
    """Test HEALPixNodes initialization."""
    node_builder = HEALPixNodes(resolution, "test_nodes")
    assert isinstance(node_builder, BaseNodeBuilder)
    assert isinstance(node_builder, HEALPixNodes)


@pytest.mark.parametrize("resolution", ["2", 4.3, -7])
def test_fail_init(resolution: int):
    """Test HEALPixNodes initialization with invalid resolution."""
    with pytest.raises(AssertionError):
        HEALPixNodes(resolution, "test_nodes")


@pytest.mark.parametrize("resolution", [2, 5, 7])
def test_register_nodes(resolution: int):
    """Test HEALPixNodes register correctly the nodes."""
    node_builder = HEALPixNodes(resolution, "test_nodes")
    graph = HeteroData()

    graph = node_builder.register_nodes(graph)

    assert graph["test_nodes"].x is not None
    assert isinstance(graph["test_nodes"].x, torch.Tensor)
    assert graph["test_nodes"].x.shape[1] == 2
    assert graph["test_nodes"].node_type == "HEALPixNodes"


@pytest.mark.parametrize("attr_class", [UniformWeights, SphericalAreaWeights])
@pytest.mark.parametrize("resolution", [2, 5, 7])
def test_register_attributes(graph_with_nodes: HeteroData, attr_class, resolution: int):
    """Test HEALPixNodes register correctly the weights."""
    node_builder = HEALPixNodes(resolution, "test_nodes")
    config = {"test_attr": {"_target_": f"anemoi.graphs.nodes.attributes.{attr_class.__name__}"}}

    graph = node_builder.register_attributes(graph_with_nodes, config)

    assert graph["test_nodes"]["test_attr"] is not None
    assert isinstance(graph["test_nodes"]["test_attr"], torch.Tensor)
    assert graph["test_nodes"]["test_attr"].shape[0] == graph["test_nodes"].x.shape[0]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import re

import torch

from anemoi.graphs.nodes.builders.base import BaseNodeBuilder
from anemoi.utils.grids import grids

LOGGER = logging.getLogger(__name__)


class ReducedGaussianGridNodes(BaseNodeBuilder):
    """Nodes from a reduced gaussian grid.

    A gaussian grid is a latitude/longitude grid. The spacing of the latitudes is not regular. However, the spacing of
    the lines of latitude is symmetrical about the Equator. A grid is usually referred to by its 'number' N/O, which
    is the number of lines of latitude between a Pole and the Equator. The N code refers to the original ECMWF reduced
    Gaussian grid, whereas the code O refers to the octahedral ECMWF reduced Gaussian grid.

    Attributes
    ----------
    grid : str
        The reduced gaussian grid, of shape {n,N,o,O}XXX with XXX latitude lines between the pole and
        equator.

    Methods
    -------
    get_coordinates()
        Get the lat-lon coordinates of the nodes.
    register_nodes(graph, name)
        Register the nodes in the graph.
    register_attributes(graph, name, config)
        Register the attributes in the nodes of the graph specified.
    update_graph(graph, name, attrs_config)
        Update the graph with new nodes and attributes.
    """

    def __init__(self, grid: int, name: str) -> None:
        """Initialize the ReducedGaussianGridNodes builder."""
        assert re.fullmatch(
            r"^[oOnN]\d+$", grid
        ), f"{self.__class__.__name__}.grid must match the format [n|N|o|O]XXX with XXX latitude lines between the pole and equator."
        self.grid = grid
        super().__init__(name)

    def get_coordinates(self) -> torch.Tensor:
        """Get the coordinates of the nodes.

        Returns
        -------
        torch.Tensor of shape (num_nodes, 2)
            A 2D tensor with the coordinates, in radians.
        """
        grid_data = grids(self.grid)
        coords = self.reshape_coords(grid_data["latitudes"], grid_data["longitudes"])
        return coords



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import h3
import networkx as nx
import numpy as np

from anemoi.graphs.generate.masks import KNNAreaMaskBuilder
from anemoi.graphs.generate.utils import get_coordinates_ordering


def create_hex_nodes(
    resolution: int,
    area_mask_builder: KNNAreaMaskBuilder | None = None,
) -> tuple[nx.Graph, np.ndarray, list[int]]:
    """Creates a global mesh from a refined icosahedron.

    This method relies on the H3 python library, which covers the earth with hexagons (and 5 pentagons). At each
    refinement level, a hexagon cell (nodes) has 7 child cells (aperture 7).

    Parameters
    ----------
    resolution : int
        Level of mesh resolution to consider.
    area_mask_builder : KNNAreaMaskBuilder, optional
        KNNAreaMaskBuilder with the cloud of points to limit the mesh area, by default None.

    Returns
    -------
    graph : networkx.Graph
        The specified graph (only nodes) sorted by latitude and longitude.
    coords_rad : np.ndarray
        The node coordinates (not ordered) in radians.
    node_ordering : list[int]
        Order of the node coordinates to be sorted by latitude and longitude.
    """
    nodes = get_nodes_at_resolution(resolution)

    coords_rad = np.deg2rad(np.array([h3.cell_to_latlng(node) for node in nodes]))

    node_ordering = get_coordinates_ordering(coords_rad)

    if area_mask_builder is not None:
        area_mask = area_mask_builder.get_mask(coords_rad)
        node_ordering = node_ordering[area_mask[node_ordering]]

    graph = create_nx_graph_from_hex_coords(nodes, node_ordering)

    return graph, coords_rad, list(node_ordering)


def create_nx_graph_from_hex_coords(nodes: list[str], node_ordering: np.ndarray) -> nx.Graph:
    """Add all nodes at a specified refinement level to a graph.

    Parameters
    ----------
    nodes : list[str]
        The set of H3 indexes (nodes).
    node_ordering: np.ndarray
        Order of the node coordinates to be sorted by latitude and longitude.

    Returns
    -------
    graph : networkx.Graph
        The graph with the added nodes.
    """
    graph = nx.Graph()

    for node_pos in node_ordering:
        h3_idx = nodes[node_pos]
        graph.add_node(h3_idx, hcoords_rad=np.deg2rad(h3.cell_to_latlng(h3_idx)))

    return graph


def get_nodes_at_resolution(
    resolution: int,
) -> list[str]:
    """Get nodes at a specified refinement level over the entire globe.

    Parameters
    ----------
    resolution : int
        The H3 refinement level. It can be an integer from 0 to 15.

    Returns
    -------
    nodes : list[str]
        The list of H3 indexes at the specified resolution level.
    """
    return list(h3.uncompact_cells(h3.get_res0_cells(), resolution))


def add_edges_to_nx_graph(
    graph: nx.Graph,
    resolutions: list[int],
    x_hops: int = 1,
    depth_children: int = 0,
) -> nx.Graph:
    """Adds the edges to the graph.

    This method includes multi-scale connections to the existing graph. The different scales
    are defined by the resolutions (or refinement levels) specified.

    Parameters
    ----------
    graph : networkx.Graph
        The graph to add the edges.
    resolutions : list[int]
        Levels of mesh resolution to consider.
    x_hops: int
        The number of hops to consider for the neighbours.
    depth_children : int
        The number of resolution levels to consider for the connections of children. Defaults to 1, which includes
        connections up to the next resolution level.

    Returns
    -------
    graph : networkx.Graph
        The graph with the added edges.
    """

    graph = add_neighbour_edges(graph, resolutions, x_hops)
    graph = add_edges_to_children(graph, resolutions, depth_children)
    return graph


def add_neighbour_edges(
    graph: nx.Graph,
    refinement_levels: tuple[int],
    x_hops: int = 1,
) -> nx.Graph:
    """Adds edges between neighbours at the specified refinement levels."""
    for resolution in refinement_levels:
        nodes = select_nodes_from_graph_at_resolution(graph, resolution)

        for idx in nodes:
            # neighbours
            for idx_neighbour in set(h3.grid_disk(idx, k=x_hops)) & set(nodes):
                graph = add_edge(
                    graph,
                    h3.cell_to_center_child(idx, max(refinement_levels)),
                    h3.cell_to_center_child(idx_neighbour, max(refinement_levels)),
                )

    return graph


def add_edges_to_children(
    graph: nx.Graph,
    refinement_levels: tuple[int],
    depth_children: int | None = None,
) -> nx.Graph:
    """Adds edges to the children of the nodes at the specified resolution levels.

    Parameters
    ----------
    graph : nx.Graph
        graph to which the edges will be added
    refinement_levels : tuple[int]
        set of refinement levels
    depth_children : int, optional
        The number of resolution levels to consider for the connections of children. Defaults to 1, which includes
        connections up to the next resolution level, by default None.

    Returns
    -------
    nx.Graph
        Graph with the added edges.
    """
    if depth_children is None:
        depth_children = len(refinement_levels)
    elif depth_children == 0:
        return graph

    for i_level, resolution_parent in enumerate(list(sorted(refinement_levels))[0:-1]):
        parent_nodes = select_nodes_from_graph_at_resolution(graph, resolution_parent)

        for parent_idx in parent_nodes:
            # add own children
            for resolution_child in refinement_levels[i_level + 1 : i_level + depth_children + 1]:
                for child_idx in h3.h3_to_children(parent_idx, res=resolution_child):
                    graph = add_edge(
                        graph,
                        h3.cell_to_center_child(parent_idx, refinement_levels[-1]),
                        h3.cell_to_center_child(child_idx, refinement_levels[-1]),
                    )

    return graph


def select_nodes_from_graph_at_resolution(graph: nx.Graph, resolution: int) -> set[str]:
    """Select nodes from a graph at a specified resolution level."""
    nodes_at_lower_resolution = [n for n in h3.compact_cells(graph.nodes) if h3.get_resolution(n) <= resolution]
    nodes_at_resolution = h3.uncompact_cells(nodes_at_lower_resolution, resolution)
    return nodes_at_resolution


def add_edge(
    graph: nx.Graph,
    source_node_h3_idx: str,
    target_node_h3_idx: str,
) -> nx.Graph:
    """Add edge between two nodes to a graph.

    The edge will only be added in case both target and source nodes are included in the graph.

    Parameters
    ----------
    graph : networkx.Graph
        The graph to add the nodes.
    source_node_h3_idx : str
        The H3 index of the tail of the edge.
    target_node_h3_idx : str
        The H3 index of the head of the edge.

    Returns
    -------
    graph : networkx.Graph
        The graph with the added edge.
    """
    if not graph.has_node(source_node_h3_idx) or not graph.has_node(target_node_h3_idx):
        return graph

    if source_node_h3_idx != target_node_h3_idx:
        graph.add_edge(source_node_h3_idx, target_node_h3_idx)

    return graph



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.nodes.attributes import SphericalAreaWeights
from anemoi.graphs.nodes.attributes import UniformWeights
from anemoi.graphs.nodes.builders.from_vectors import LatLonNodes

lats = [45.0, 45.0, 40.0, 40.0]
lons = [5.0, 10.0, 10.0, 5.0]


def test_init():
    """Test LatLonNodes initialization."""
    node_builder = LatLonNodes(latitudes=lats, longitudes=lons, name="test_nodes")
    assert isinstance(node_builder, LatLonNodes)


def test_fail_init_length_mismatch():
    """Test LatLonNodes initialization with invalid argument."""
    lons = [5.0, 10.0, 10.0, 5.0, 5.0]

    with pytest.raises(AssertionError):
        LatLonNodes(latitudes=lats, longitudes=lons, name="test_nodes")


def test_fail_init_missing_argument():
    """Test NPZFileNodes initialization with missing argument."""
    with pytest.raises(TypeError):
        LatLonNodes(name="test_nodes")


def test_register_nodes():
    """Test LatLonNodes register correctly the nodes."""
    graph = HeteroData()
    node_builder = LatLonNodes(latitudes=lats, longitudes=lons, name="test_nodes")
    graph = node_builder.register_nodes(graph)

    assert graph["test_nodes"].x is not None
    assert isinstance(graph["test_nodes"].x, torch.Tensor)
    assert graph["test_nodes"].x.shape == (len(lats), 2)
    assert graph["test_nodes"].node_type == "LatLonNodes"


@pytest.mark.parametrize("attr_class", [UniformWeights, SphericalAreaWeights])
def test_register_attributes(graph_with_nodes: HeteroData, attr_class):
    """Test LatLonNodes register correctly the weights."""
    node_builder = LatLonNodes(latitudes=lats, longitudes=lons, name="test_nodes")
    config = {"test_attr": {"_target_": f"anemoi.graphs.nodes.attributes.{attr_class.__name__}"}}

    graph = node_builder.register_attributes(graph_with_nodes, config)

    assert graph["test_nodes"]["test_attr"] is not None
    assert isinstance(graph["test_nodes"]["test_attr"], torch.Tensor)
    assert graph["test_nodes"]["test_attr"].shape[0] == graph["test_nodes"].x.shape[0]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import numpy as np
import pytest
import torch
import xarray as xr
import yaml
from nodes.test_icon_nodes import DatasetMock
from torch_geometric.data import HeteroData

lats = [-0.15, 0, 0.15]
lons = [0, 0.25, 0.5, 0.75]


class MockAnemoiDataset:
    """Mock Zarr dataset with latitudes and longitudes attributes."""

    def __init__(self, latitudes, longitudes, grids=None):
        self.latitudes = latitudes
        self.longitudes = longitudes
        self.num_nodes = len(latitudes)
        self.grids = grids if grids is not None else (self.num_nodes,)


@pytest.fixture
def mock_anemoi_dataset() -> MockAnemoiDataset:
    """Mock zarr dataset with nodes."""
    coords = 2 * torch.pi * np.array([[lat, lon] for lat in lats for lon in lons])
    return MockAnemoiDataset(latitudes=coords[:, 0], longitudes=coords[:, 1])


@pytest.fixture
def mock_zarr_dataset_file(tmpdir) -> str:
    lat_vals = np.linspace(-90, 90, 5)
    lon_vals = np.linspace(0, 360, 5, endpoint=False)
    lat, lon = np.meshgrid(lat_vals, lon_vals, indexing="ij")
    data = np.random.randn(5, 5)

    ds = xr.Dataset(
        {
            "variable": (
                ["lat", "lon"],
                data,
            ),
        },
        coords={
            "lat": (["x", "y"], lat),
            "lon": (["x", "y"], lon),
        },
    )

    fn = tmpdir / "tmp.zarr"
    ds.to_zarr(fn, mode="w", consolidated=True)

    return fn


@pytest.fixture
def mock_anemoi_dataset_cutout() -> MockAnemoiDataset:
    """Mock zarr dataset with nodes."""
    coords = 2 * torch.pi * np.array([[lat, lon] for lat in lats for lon in lons])
    grids = int(0.3 * len(coords)), len(coords) - int(0.3 * len(coords))
    return MockAnemoiDataset(latitudes=coords[:, 0], longitudes=coords[:, 1], grids=grids)


@pytest.fixture
def mock_grids_path(tmp_path) -> tuple[str, int]:
    """Mock grid_definition_path with files for 3 resolutions."""
    num_nodes = len(lats) * len(lons)
    for resolution in ["o16", "o48", "5km5"]:
        file_path = tmp_path / f"grid-{resolution}.npz"
        np.savez(
            file_path,
            latitudes=np.random.rand(num_nodes),
            longitudes=np.random.rand(num_nodes),
        )
    return str(tmp_path), num_nodes


@pytest.fixture
def graph_with_nodes() -> HeteroData:
    """Graph with 12 nodes over the globe, stored in \"test_nodes\"."""
    coords = np.array([[lat, lon] for lat in lats for lon in lons])
    graph = HeteroData()
    graph["test_nodes"].x = 2 * torch.pi * torch.tensor(coords)
    graph["test_nodes"].mask = torch.tensor([True] * len(coords)).unsqueeze(-1)
    graph["test_nodes"].mask2 = torch.tensor([True] * (len(coords) - 2) + [False] * 2).unsqueeze(-1)
    graph["test_nodes"].interior_mask = torch.tensor(
        [
            False,
            False,
            False,
            False,
            False,
            True,
            True,
            False,
            False,
            False,
            False,
            False,
        ]
    ).unsqueeze(-1)
    graph["test_nodes"]["_grid_reference_distance"] = 0.75
    return graph


@pytest.fixture
def graph_with_rectilinear_nodes() -> HeteroData:
    graph = HeteroData()
    num_lons, num_lats = 10, 10
    lat_grid, lon_grid = np.meshgrid(np.linspace(-np.pi / 2, np.pi / 2, num_lats), np.linspace(0, 2 * np.pi, num_lons))
    coords = torch.tensor(np.array([lat_grid.ravel(), lon_grid.ravel()]).T)
    graph["test_nodes"].x = coords
    return graph


@pytest.fixture
def graph_with_isolated_nodes() -> HeteroData:
    graph = HeteroData()
    graph["test_nodes"].x = torch.tensor([[1], [2], [3], [4], [5], [6]])
    graph["test_nodes"]["mask_attr"] = torch.tensor([[1], [1], [1], [0], [0], [0]], dtype=torch.bool)
    graph["test_nodes", "to", "test_nodes"].edge_index = torch.tensor([[2, 3, 4], [1, 2, 3]])
    return graph


@pytest.fixture
def graph_nodes_and_edges() -> HeteroData:
    """Graph with 1 set of nodes and edges."""
    coords = np.array([[lat, lon] for lat in lats for lon in lons])
    graph = HeteroData()
    graph["test_nodes"].x = 2 * torch.pi * torch.tensor(coords)
    graph["test_nodes"].mask = torch.tensor([True] * len(coords)).unsqueeze(-1)
    graph[("test_nodes", "to", "test_nodes")].edge_index = torch.tensor([[3, 1, 2, 0], [2, 0, 1, 3]])
    graph[("test_nodes", "to", "test_nodes")].edge_attr = (
        10 * graph[("test_nodes", "to", "test_nodes")].edge_index[0][:, None]
    )
    return graph


@pytest.fixture
def graph_long_and_short_edges() -> HeteroData:
    """Graph with a pair of short (800km) and a pair of long (20000km) edges."""
    graph = HeteroData()
    graph["test_nodes"].x = 2 * torch.pi * torch.tensor([[-0.01, 0], [0.01, 0], [-0.01, 0.5], [0.01, 0.5]])
    graph["test_nodes"]["southern_hemisphere_mask"] = torch.tensor([[1], [0], [1], [0]], dtype=torch.bool)
    graph["test_nodes", "to", "test_nodes"].edge_index = torch.tensor([[0, 0, 1, 3], [1, 3, 2, 2]])
    return graph


@pytest.fixture(scope="session")
def icon_dataset_mock():
    """AICON graph dataset mock."""
    return DatasetMock


@pytest.fixture
def config_file(tmp_path) -> tuple[str, str]:
    """Mock grid_definition_path with files for 3 resolutions."""
    cfg = {
        "nodes": {
            "test_nodes": {
                "node_builder": {
                    "_target_": "anemoi.graphs.nodes.NPZFileNodes",
                    "npz_file": str(tmp_path) + "/grid-o16.npz",
                },
            },
        },
        "edges": [
            {
                "source_name": "test_nodes",
                "target_name": "test_nodes",
                "edge_builders": [
                    {
                        "_target_": "anemoi.graphs.edges.KNNEdges",
                        "num_nearest_neighbours": 3,
                    },
                ],
                "attributes": {
                    "dist_norm": {"_target_": "anemoi.graphs.edges.attributes.EdgeLength"},
                    "edge_dirs": {"_target_": "anemoi.graphs.edges.attributes.EdgeDirection"},
                },
            },
        ],
    }
    file_name = "config.yaml"

    with (tmp_path / file_name).open("w") as file:
        yaml.dump(cfg, file)

    return tmp_path, file_name



# (C) Copyright 2024- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.nodes.attributes import BooleanAndMask
from anemoi.graphs.nodes.attributes import BooleanNot
from anemoi.graphs.nodes.attributes import BooleanOrMask


def test_boolean_not(graph_with_nodes: HeteroData):
    """Test attribute builder for BooleanNot."""
    node_attr_builder = BooleanNot("mask")
    mask = node_attr_builder.compute(graph_with_nodes, "test_nodes")

    assert mask is not None
    assert isinstance(mask, torch.Tensor)
    assert mask.dtype == torch.bool
    assert mask.shape[0] == graph_with_nodes["test_nodes"].x.shape[0]


def test_boolean_fail_multiple_masks(graph_with_nodes: HeteroData):
    """Test attribute builder for BooleanNot."""
    node_attr_builder = BooleanNot(["mask", "mask2"])
    with pytest.raises(AssertionError):
        node_attr_builder.compute(graph_with_nodes, "test_nodes")


def test_boolean_and_mask(graph_with_nodes: HeteroData):
    """Test attribute builder for BooleanAndMask."""
    node_attr_builder = BooleanAndMask(["mask2"])
    mask = node_attr_builder.compute(graph_with_nodes, "test_nodes")
    assert torch.allclose(mask, graph_with_nodes["test_nodes"]["mask2"])

    node_attr_builder = BooleanAndMask(["mask", "mask2"])
    mask = node_attr_builder.compute(graph_with_nodes, "test_nodes")

    assert mask is not None
    assert isinstance(mask, torch.Tensor)
    assert mask.dtype == torch.bool
    assert mask.shape[0] == graph_with_nodes["test_nodes"].x.shape[0]


def test_boolean_or_mask(graph_with_nodes: HeteroData):
    """Test attribute builder for BooleanOrMask."""
    node_attr_builder = BooleanOrMask(["mask2"])
    mask = node_attr_builder.compute(graph_with_nodes, "test_nodes")
    print(mask.shape)
    print(graph_with_nodes["test_nodes"]["mask2"].shape)
    assert torch.allclose(mask, graph_with_nodes["test_nodes"]["mask2"])

    node_attr_builder = BooleanOrMask(["mask", "mask2"])
    mask = node_attr_builder.compute(graph_with_nodes, "test_nodes")

    assert mask is not None
    assert isinstance(mask, torch.Tensor)
    assert mask.dtype == torch.bool
    assert mask.shape[0] == graph_with_nodes["test_nodes"].x.shape[0]


def test_wrong_mask(graph_with_nodes: HeteroData):
    node_attr_builder = BooleanOrMask(["mask", "askdbash]"])
    with pytest.raises(AssertionError):
        node_attr_builder.compute(graph_with_nodes, "test_nodes")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import netCDF4
import numpy as np
import pytest
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.generate.icon_mesh import ICONCellDataGrid
from anemoi.graphs.generate.icon_mesh import ICONMultiMesh
from anemoi.graphs.nodes import ICONCellGridNodes
from anemoi.graphs.nodes import ICONMultiMeshNodes
from anemoi.graphs.nodes.builders.base import BaseNodeBuilder


class DatasetMock:
    """This datasets emulates the most primitive unstructured grid with
    refinement.

    Enumeration of cells , edges and vertices in netCDF file is 1 based.
    C: cell
    E: edge
    V: vertex

    Cell C2 with its additional vertex V4 and edges E4 and E4 were added as
    a first refinement.

    [V1: 0, 1]🢀-E3--[V3: 1, 1]
      🢁      ╲             🢁
      |       ╲ [C1: ⅔, ⅔] |
      |        ╲           |
      E5        E1         E2
      |          ╲         |
      |           ╲        |
      | [C2: ⅓, ⅓] ╲       |
      |             🢆     |
    [V4: 0, 1]🢀-E4--[V2: 1, 1]

    Note: Triangular refinement does not actually work like this. This grid
    mock serves testing purposes only.

    """

    def __init__(self, *args, **kwargs):

        class MockVariable:
            def __init__(self, data, units, dimensions):
                self.data = np.ma.asarray(data)
                self.shape = data.shape
                self.units = units
                self.dimensions = dimensions

            def __getitem__(self, key):
                return self.data[key]

        self.variables = {
            "vlon": MockVariable(np.array([0, 1, 1, 0]), "radian", ("vertex",)),
            "vlat": MockVariable(np.array([1, 0, 1, 0]), "radian", ("vertex",)),
            "clon": MockVariable(np.array([0.66, 0.33]), "radian", ("cell",)),
            "clat": MockVariable(np.array([0.66, 0.33]), "radian", ("cell",)),
            "edge_vertices": MockVariable(np.array([[1, 2], [2, 3], [3, 1], [2, 4], [4, 1]]).T, "", ("nc", "edge")),
            "vertex_of_cell": MockVariable(np.array([[1, 2, 3], [1, 2, 4]]).T, "", ("nv", "cell")),
            "refinement_level_v": MockVariable(np.array([0, 0, 0, 1]), "", ("vertex",)),
            "refinement_level_c": MockVariable(np.array([0, 1]), "", ("cell",)),
        }
        """common array dimensions:
            nc: 2, # constant
            nv: 3, # constant
            vertex: 4,
            edge: 5,
            cell: 2,
        """
        self.uuidOfHGrid = "__test_data__"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


@pytest.mark.parametrize("max_level", [0, 1, 2])
@pytest.mark.parametrize("node_builder_cls", [ICONMultiMeshNodes, ICONCellGridNodes])
def test_init(monkeypatch, max_level: int, node_builder_cls: type[BaseNodeBuilder]):
    """Test ICON node builders initialization."""

    monkeypatch.setattr(netCDF4, "Dataset", DatasetMock)
    node_builder = node_builder_cls(
        name="test_nodes",
        grid_filename="test.nc",
        max_level=max_level,
    )
    assert isinstance(node_builder, BaseNodeBuilder)
    assert hasattr(node_builder, "icon_nodes")
    assert isinstance(node_builder.icon_nodes, (ICONMultiMesh, ICONCellDataGrid))
    assert hasattr(node_builder.icon_nodes, "nodeset")
    assert hasattr(node_builder.icon_nodes.nodeset, "gc_vertices")


@pytest.mark.parametrize("node_builder_cls", [ICONCellGridNodes, ICONMultiMeshNodes])
def test_node_builder_dependencies(monkeypatch, node_builder_cls: type[BaseNodeBuilder]):
    """Test that the `node_builder` depends on the presence of ICON node builders."""
    monkeypatch.setattr(netCDF4, "Dataset", DatasetMock)
    node_builder = node_builder_cls(name="data_nodes", max_level=0, grid_filename="test.nc")

    graph = HeteroData()
    graph = node_builder.update_graph(graph)

    assert isinstance(graph, HeteroData)
    assert "data_nodes" in graph.node_types


@pytest.mark.parametrize("node_builder_cls", [ICONCellGridNodes, ICONMultiMeshNodes])
def test_wrong_filename(node_builder_cls: type[BaseNodeBuilder]):
    with pytest.raises(FileNotFoundError):
        node_builder_cls(name="data_nodes2", max_level=0, grid_filename="missing_icon_nodes")


def test_register_nodes(monkeypatch):
    """Test ICON node builders register correctly the nodes."""
    monkeypatch.setattr(netCDF4, "Dataset", DatasetMock)

    node_builder = ICONMultiMeshNodes(name="test_icon_nodes", grid_filename="test.nc", max_level=0)

    graph = node_builder.register_nodes(HeteroData())

    assert graph["test_icon_nodes"].x is not None
    assert isinstance(graph["test_icon_nodes"].x, torch.Tensor)
    assert graph["test_icon_nodes"].x.shape[1] == 2
    assert graph["test_icon_nodes"].num_nodes == 3, "number of vertices at refinement_level_v == 0"
    assert graph["test_icon_nodes"].node_type == "ICONMultiMeshNodes"

    node_builder2 = ICONMultiMeshNodes(name="test_icon_nodes", grid_filename="test.nc", max_level=1)
    graph = node_builder2.register_nodes(HeteroData())
    assert graph["test_icon_nodes"].num_nodes == 4, "number of vertices at refinement_level_v == 1"
    assert graph["test_icon_nodes"].node_type == "ICONMultiMeshNodes"


def test_register_attributes(
    monkeypatch,
    graph_with_nodes: HeteroData,
):
    """Test ICONNodes register correctly the weights."""
    monkeypatch.setattr(netCDF4, "Dataset", DatasetMock)
    nodes = ICONCellGridNodes(name="test_nodes", max_level=0, grid_filename="test.nc")
    config = {"test_attr": {"_target_": "anemoi.graphs.nodes.attributes.UniformWeights"}}

    graph = nodes.register_attributes(graph_with_nodes, config)

    assert "test_attr" in graph["test_nodes"]
    assert torch.mean(graph["test_nodes"].test_attr) == 1.0
    assert isinstance(graph["test_nodes"]["_icon_nodes"], ICONCellDataGrid)
    assert hasattr(graph["test_nodes"]["_icon_nodes"], "grid_filename")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from abc import ABC
from abc import abstractmethod

import networkx as nx
import numpy as np
from torch_geometric.data.storage import NodeStorage

LOGGER = logging.getLogger(__name__)


class BaseIcosahedronEdgeStrategy(ABC):
    """Abstract base class for different edge-building strategies."""

    @abstractmethod
    def get_edges(self, nodes: NodeStorage, x_hops: int, scale_resolutions: list[int]) -> NodeStorage: ...


class TriNodesEdgeBuilder(BaseIcosahedronEdgeStrategy):
    """Edge builder for TriNodes and LimitedAreaTriNodes."""

    def get_edges(
        self,
        nodes: NodeStorage,
        x_hops: int,
        scale_resolutions: list[int],
    ) -> NodeStorage:
        from anemoi.graphs.generate import tri_icosahedron

        if x_hops == 1:
            LOGGER.debug("The 1-hop edges are obtained directly from the trimesh.Trimesh object's 'edges' attribute.")
            # Compute the multiscale edges directly and store them in the node storage
            # No need of the networkx graph
            multiscale_edges = tri_icosahedron.add_1_hop_edges(
                nodes_coords_rad=nodes["x"],
                node_resolutions=nodes["_resolutions"],
                edge_resolutions=scale_resolutions,
                node_ordering=nodes["_node_ordering"],
                area_mask_builder=nodes.get("_area_mask_builder", None),
            )

        else:
            LOGGER.info("Using networkx strategy for multiscale-edge building.")
            nx_graph = tri_icosahedron.add_edges_to_nx_graph(
                nodes["_nx_graph"],
                resolutions=scale_resolutions,
                x_hops=x_hops,
                area_mask_builder=nodes.get("_area_mask_builder", None),
            )
            adjmat = nx.to_scipy_sparse_array(nx_graph, format="coo")
            # Get source & target indices of the edges
            multiscale_edges = np.stack([adjmat.col, adjmat.row], axis=0)
        return multiscale_edges


class HexNodesEdgeBuilder(BaseIcosahedronEdgeStrategy):
    """Edge builder for HexNodes and LimitedAreaHexNodes."""

    def get_edges(self, nodes: NodeStorage, x_hops: int, scale_resolutions: list[int]) -> NodeStorage:
        from anemoi.graphs.generate import hex_icosahedron

        nx_graph = hex_icosahedron.add_edges_to_nx_graph(
            nodes["_nx_graph"],
            resolutions=scale_resolutions,
            x_hops=x_hops,
        )
        adjmat = nx.to_scipy_sparse_array(nx_graph, format="coo")
        # Get source & target indices of the edges
        multiscale_edges = np.stack([adjmat.col, adjmat.row], axis=0)

        return multiscale_edges


class StretchedTriNodesEdgeBuilder(BaseIcosahedronEdgeStrategy):
    """Edge builder for StretchedTriNodes."""

    def get_edges(self, nodes: NodeStorage, x_hops: int, scale_resolutions: list[int]) -> NodeStorage:
        from anemoi.graphs.generate import tri_icosahedron
        from anemoi.graphs.generate.masks import KNNAreaMaskBuilder

        all_points_mask_builder = KNNAreaMaskBuilder("all_nodes", 1.0)
        all_points_mask_builder.fit_coords(nodes.x.cpu().numpy())

        if x_hops == 1:
            LOGGER.debug("Using tri-mesh only strategy for x_hops=1 multiscale-edge building.")
            # Compute the multiscale edges directly and store them in the node storage
            # No need of the networkx graph
            multiscale_edges = tri_icosahedron.add_1_hop_edges(
                nodes_coords_rad=nodes["x"],
                node_resolutions=nodes["_resolutions"],
                edge_resolutions=scale_resolutions,
                node_ordering=nodes["_node_ordering"],
                area_mask_builder=all_points_mask_builder,
            )

        else:
            nx_graph = tri_icosahedron.add_edges_to_nx_graph(
                nodes["_nx_graph"],
                resolutions=scale_resolutions,
                x_hops=x_hops,
                area_mask_builder=all_points_mask_builder,
            )
            adjmat = nx.to_scipy_sparse_array(nx_graph, format="coo")
            # Get source & target indices of the edges
            multiscale_edges = np.stack([adjmat.col, adjmat.row], axis=0)

        return multiscale_edges



# (C) Copyright 2024- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import pytest
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.nodes.attributes import CosineLatWeightedAttribute
from anemoi.graphs.nodes.attributes import IsolatitudeAreaWeights
from anemoi.graphs.nodes.attributes import MaskedPlanarAreaWeights
from anemoi.graphs.nodes.attributes import PlanarAreaWeights
from anemoi.graphs.nodes.attributes import SphericalAreaWeights
from anemoi.graphs.nodes.attributes import UniformWeights
from anemoi.graphs.nodes.attributes.base_attributes import BaseNodeAttribute


def test_uniform_weights(graph_with_nodes: HeteroData):
    """Test attribute builder for UniformWeights."""
    node_attr_builder = UniformWeights()
    weights = node_attr_builder.compute(graph_with_nodes, "test_nodes")

    # All values must be the same. Then, the mean has to be also the same
    assert torch.max(torch.abs(weights - torch.mean(weights))) == 0
    assert isinstance(weights, torch.Tensor)
    assert weights.shape[0] == graph_with_nodes["test_nodes"].x.shape[0]
    assert weights.dtype == node_attr_builder.dtype


def test_planar_area_weights(graph_with_nodes: HeteroData):
    """Test attribute builder for PlanarAreaWeights."""
    node_attr_builder = PlanarAreaWeights()
    weights = node_attr_builder.compute(graph_with_nodes, "test_nodes")

    assert weights is not None
    assert isinstance(weights, torch.Tensor)
    assert weights.shape[0] == graph_with_nodes["test_nodes"].x.shape[0]
    assert weights.dtype == node_attr_builder.dtype


@pytest.mark.parametrize("fill_value", [0.0, -1.0, float("nan")])
def test_spherical_area_weights(graph_with_nodes: HeteroData, fill_value: float):
    """Test attribute builder for SphericalAreaWeights with different fill values."""
    node_attr_builder = SphericalAreaWeights(fill_value=fill_value)
    weights = node_attr_builder.compute(graph_with_nodes, "test_nodes")

    assert weights is not None
    assert isinstance(weights, torch.Tensor)
    assert weights.shape[0] == graph_with_nodes["test_nodes"].x.shape[0]
    assert weights.dtype == node_attr_builder.dtype


@pytest.mark.parametrize("radius", [-1.0, "hello", None])
def test_spherical_area_weights_wrong_radius(radius: float):
    """Test attribute builder for SphericalAreaWeights with invalid radius."""
    with pytest.raises(AssertionError):
        SphericalAreaWeights(radius=radius)


@pytest.mark.parametrize("fill_value", ["invalid", "as"])
def test_spherical_area_weights_wrong_fill_value(fill_value: str):
    """Test attribute builder for SphericalAreaWeights with invalid fill_value."""
    with pytest.raises(AssertionError):
        SphericalAreaWeights(fill_value=fill_value)


@pytest.mark.parametrize("attr_class", [IsolatitudeAreaWeights, CosineLatWeightedAttribute])
@pytest.mark.parametrize("norm", [None, "l1", "unit-max"])
def test_latweighted(attr_class: type[BaseNodeAttribute], graph_with_rectilinear_nodes, norm: str):
    """Test attribute builder for Lat with different fill values."""
    node_attr_builder = attr_class(norm=norm)
    weights = node_attr_builder.compute(graph_with_rectilinear_nodes, "test_nodes")

    assert weights is not None
    assert isinstance(weights, torch.Tensor)
    assert torch.all(weights >= 0)
    assert weights.shape[0] == graph_with_rectilinear_nodes["test_nodes"].x.shape[0]
    assert weights.dtype == node_attr_builder.dtype


def test_masked_planar_area_weights(graph_with_nodes: HeteroData):
    """Test attribute builder for PlanarAreaWeights."""
    node_attr_builder = MaskedPlanarAreaWeights(mask_node_attr_name="interior_mask")
    weights = node_attr_builder.compute(graph_with_nodes, "test_nodes")

    assert weights is not None
    assert isinstance(weights, torch.Tensor)
    assert weights.shape[0] == graph_with_nodes["test_nodes"].x.shape[0]
    assert weights.dtype == node_attr_builder.dtype

    mask = graph_with_nodes["test_nodes"]["interior_mask"]
    assert torch.all(weights[~mask] == 0)


def test_masked_planar_area_weights_fail(graph_with_nodes: HeteroData):
    """Test attribute builder for AreaWeights with invalid radius."""
    with pytest.raises(AssertionError):
        node_attr_builder = MaskedPlanarAreaWeights(mask_node_attr_name="nonexisting")
        node_attr_builder.compute(graph_with_nodes, "test_nodes")



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging
from typing import Annotated
from typing import Iterable
from typing import Literal

from pydantic import Field

from anemoi.utils.schemas import BaseModel

LOGGER = logging.getLogger(__name__)


class RemoveUnconnectedNodesSchema(BaseModel):
    target_: Literal["anemoi.graphs.processors.RemoveUnconnectedNodes"] = Field(..., alias="_target_")
    "Post processor to remove unconnected nodes."
    nodes_name: str | Iterable[str]
    "Nodes from which to remove the unconnected nodes."
    ignore: str = Field(example=None)
    "Attribute name of nodes to be ignored."
    save_mask_indices_to_attr: str = Field(example=None)
    "New attribute name to store the mask indices."


class SubsetNodesInAreaSchema(BaseModel):
    target_: Literal["anemoi.graphs.processors.SubsetNodesInArea"] = Field(..., alias="_target_")
    "Post processor to remove unconnected nodes."
    nodes_name: str | Iterable[str]
    "Nodes from which to remove the unconnected nodes."
    area: tuple[float, float, float, float] = Field(default=(40, 10, 30, 20))
    "Area of interest to crop the nodes, (north, west, south, east)."
    save_mask_indices_to_attr: str = Field(example=None)
    "New attribute name to store the mask indices."


class RestrictEdgeLengthSchema(BaseModel):
    target_: Literal["anemoi.graphs.processors.RestrictEdgeLength"] = Field(..., alias="_target_")
    "Post processor to edges longer than a threshold."
    source_name: str
    "Source nodes of edges to be post-processed."
    target_name: str
    "Target nodes of edges to be post-processed."
    max_length_km: float
    "Treshold length (in km), edges longer than this length will be removed"
    source_mask_attr_name: str | None = Field(default=None, example=None)
    "Boolean mask attribute on sources nodes. Only edges whose source is True under this mask will be post-processed. Default to None"
    target_mask_attr_name: str | None = Field(default=None, example=None)
    "Boolean mask attribute on target nodes. Only edges whose target is True under this mask will be post-processed. Default to None"


class SortEdgeIndexSchema(BaseModel):
    target_: Literal[
        "anemoi.graphs.processors.SortEdgeIndexBySourceNodes",
        "anemoi.graphs.processors.SortEdgeIndexByTargetNodes",
    ] = Field(..., alias="_target_")
    "Post processor to sort edge indices based on either source or target nodes."
    descending: bool = Field(default=True, example=True)
    "Flag to sort edge indices in descending order."


ProcessorSchemas = Annotated[
    RemoveUnconnectedNodesSchema | SubsetNodesInAreaSchema | RestrictEdgeLengthSchema | SortEdgeIndexSchema,
    Field(discriminator="target_"),
]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

import numpy as np
from sklearn.neighbors import NearestNeighbors
from torch_geometric.data import HeteroData

from anemoi.graphs import EARTH_RADIUS

LOGGER = logging.getLogger(__name__)


class KNNAreaMaskBuilder:
    """Class to build a mask based on distance to masked reference nodes using KNN.

    Attributes
    ----------
    nearest_neighbour : NearestNeighbors
        Nearest neighbour object to compute the KNN.
    margin_radius_km : float
        Maximum distance to the reference nodes to consider a node as valid, in kilometers. Defaults to 100 km.
    reference_node_name : str
        Name of the reference nodes in the graph to consider for the Area Mask.
    mask_attr_name : str
        Name of a node to attribute to mask the reference nodes, if desired. Defaults to consider all reference nodes.

    Methods
    -------
    fit_coords(coords_rad: np.ndarray)
        Fit the KNN model to the coordinates in radians.
    fit(graph: HeteroData)
        Fit the KNN model to the reference nodes.
    get_mask(coords_rad: np.ndarray) -> np.ndarray
        Get the mask for the nodes based on the distance to the reference nodes.
    """

    def __init__(self, reference_node_name: str, margin_radius_km: float = 100, mask_attr_name: str | None = None):
        assert isinstance(margin_radius_km, (int, float)), "The margin radius must be a number."
        assert margin_radius_km > 0, "The margin radius must be positive."

        self.nearest_neighbour = NearestNeighbors(metric="haversine", n_jobs=4)
        self.margin_radius_km = margin_radius_km
        self.reference_node_name = reference_node_name
        self.mask_attr_name = mask_attr_name

    def get_reference_coords(self, graph: HeteroData) -> np.ndarray:
        """Retrive coordinates from the reference nodes."""
        assert (
            self.reference_node_name in graph.node_types
        ), f'Reference node "{self.reference_node_name}" not found in the graph.'

        coords_rad = graph[self.reference_node_name].x
        if self.mask_attr_name is not None:
            assert (
                self.mask_attr_name in graph[self.reference_node_name].node_attrs()
            ), f'Mask attribute "{self.mask_attr_name}" not found in the reference nodes.'
            mask = graph[self.reference_node_name][self.mask_attr_name].squeeze()
            coords_rad = coords_rad[mask]

        return coords_rad.cpu().numpy()

    def fit_coords(self, coords_rad: np.ndarray):
        """Fit the KNN model to the coordinates in radians."""
        self.nearest_neighbour.fit(coords_rad)

    def fit(self, graph: HeteroData):
        """Fit the KNN model to the nodes of interest."""
        # Prepare string for logging
        reference_mask_str = self.reference_node_name
        if self.mask_attr_name is not None:
            reference_mask_str += f" ({self.mask_attr_name})"

        # Fit to the reference nodes
        coords_rad = self.get_reference_coords(graph)
        self.fit_coords(coords_rad)

        LOGGER.info(
            'Fitting %s with %d reference nodes from "%s".',
            self.__class__.__name__,
            len(coords_rad),
            reference_mask_str,
        )

    def get_mask(self, coords_rad: np.ndarray) -> np.ndarray:
        """Compute a mask based on the distance to the reference nodes."""

        neigh_dists, _ = self.nearest_neighbour.kneighbors(coords_rad, n_neighbors=1)
        mask = neigh_dists[:, 0] * EARTH_RADIUS <= self.margin_radius_km
        return mask



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import torch

from anemoi.graphs.generate.transforms import latlon_rad_to_sincos
from anemoi.graphs.generate.transforms import sincos_to_latlon_rad

NUM_POINTS = 40


def test_latlon_rad_to_sincos():
    """Test conversion from lat-lon in radians to sin-cos representation."""
    coords = torch.rand(NUM_POINTS, 2) * torch.tensor([torch.pi, 2 * torch.pi]) - torch.tensor([torch.pi / 2, 0])
    # lat in [-pi/2, pi/2], lon in [0, 2 * pi]

    sincos = latlon_rad_to_sincos(coords)
    assert sincos.shape == (NUM_POINTS, 4)
    assert sincos.dtype == coords.dtype
    assert sincos.device == coords.device
    assert sincos.requires_grad == coords.requires_grad

    assert torch.all((sincos[:, :2] >= -1) & (sincos[:, :2] <= 1))  # sin(lat), sin(lon)
    assert torch.all((sincos[:, 2] >= 0) & (sincos[:, 2] <= 1))  # cos(lat)
    assert torch.all((sincos[:, 3] >= -1) & (sincos[:, 3] <= 1))  # cos(lon)


def test_sincos_to_latlon_rad():
    """Test conversion from sin-cos representation to lat-lon in radians."""
    sincos = torch.rand(NUM_POINTS, 4) * torch.tensor([2, 2, 1, 2]) - torch.tensor([1, 1, 0, 1])
    # sin(lat), sin(lon), cos(lon) in [-1, 1], cos(lat) in [0, 1]

    coords = sincos_to_latlon_rad(sincos)
    assert coords.shape == (NUM_POINTS, 2)
    assert coords.dtype == sincos.dtype
    assert coords.device == sincos.device
    assert coords.requires_grad == sincos.requires_grad

    assert torch.all((coords[:, 0] >= -torch.pi / 2) & (coords[:, 0] <= torch.pi / 2))  # lat in [-pi/2, pi/2]
    assert torch.all((coords[:, 1] >= 0) & (coords[:, 1] <= 2 * torch.pi))  # lon in [0, 2 * pi]


def test_latlon_rad_to_sincos_and_back():
    """Test conversion from lat-lon in radians to sin-cos representation and back."""
    coords = torch.rand(NUM_POINTS, 2) * torch.tensor([torch.pi, 2 * torch.pi]) - torch.tensor([torch.pi / 2, 0])

    sincos = latlon_rad_to_sincos(coords)
    assert sincos.shape == (NUM_POINTS, 4)

    recovered = sincos_to_latlon_rad(sincos)
    assert recovered.shape == (NUM_POINTS, 2)

    torch.testing.assert_close(recovered, coords)



from .post_process import RemoveUnconnectedNodes
from .post_process import RestrictEdgeLength
from .post_process import SortEdgeIndexBySourceNodes
from .post_process import SortEdgeIndexByTargetNodes

__all__ = [
    "RemoveUnconnectedNodes",
    "RestrictEdgeLength",
    "SortEdgeIndexByTargetNodes",
    "SortEdgeIndexBySourceNodes",
]



# (C) Copyright 2024- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.nodes.attributes import CutOutMask
from anemoi.graphs.nodes.attributes import GridsMask
from anemoi.graphs.nodes.attributes.masks import BaseCombineAnemoiDatasetsMask


def test_cutout_mask(mocker, graph_with_nodes: HeteroData, mock_anemoi_dataset_cutout):
    """Test attribute builder for CutOutMask."""
    # Add dataset attribute required by CutOutMask
    graph_with_nodes["test_nodes"]["_dataset"] = {}

    mocker.patch("anemoi.datasets.open_dataset", return_value=mock_anemoi_dataset_cutout)
    mask = CutOutMask().compute(graph_with_nodes, "test_nodes")

    assert mask is not None
    assert isinstance(mask, torch.Tensor)
    assert mask.dtype == torch.bool
    assert mask.shape[0] == graph_with_nodes["test_nodes"].x.shape[0]


def test_get_mask_from_grid_size():
    grid_sizes1 = (1, 2, 3, 2, 1)
    grid_sizes2 = (4,)
    grid_ids1 = [0, 1, 4]
    grid_ids2 = [1]

    grids_mask1 = BaseCombineAnemoiDatasetsMask.get_mask_from_grid_sizes(grid_sizes1, grid_ids1)
    grids_mask2 = BaseCombineAnemoiDatasetsMask.get_mask_from_grid_sizes(grid_sizes1, grid_ids2)

    with pytest.raises(AssertionError):
        BaseCombineAnemoiDatasetsMask.get_mask_from_grid_sizes(grid_sizes2, grid_ids2)

    assert all(grids_mask1 == torch.tensor([1, 1, 1, 0, 0, 0, 0, 0, 1], dtype=torch.bool))
    assert all(grids_mask2 == torch.tensor([0, 1, 1, 0, 0, 0, 0, 0, 0], dtype=torch.bool))


@pytest.mark.parametrize("mask_class", [CutOutMask, GridsMask])
def test_combined_datasets_mask_missing_dataset(graph_with_nodes: HeteroData, mask_class):
    """Test CutOutMask fails when dataset attribute is missing."""
    node_attr_builder = mask_class()
    with pytest.raises(AssertionError):
        node_attr_builder.compute(graph_with_nodes, "test_nodes")


if __name__ == "__main__":
    pytest.main([__file__])



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import torch

from anemoi.graphs.generate.transforms import latlon_rad_to_cartesian

NORTH_POLE = [0.0, 0.0, 1.0]  # North pole in 3D coordinates


def rotate_vectors(v: torch.Tensor, axis: torch.Tensor, angle: torch.Tensor, eps=1e-8) -> torch.Tensor:
    """Rotate points v around axis by the angle using Rodrigues' rotation formula in torch.

    Parameters
    ----------
    v : torch.Tensor
        A tensor of shape (N, 3) representing N vectors to be rotated.
    axis : torch.Tensor
        A tensor of shape (N, 3) representing the rotation axis.
    angle : torch.Tensor
        A tensor of shape (N,) representing the rotation angles.

    Returns
    -------
    torch.Tensor
        A tensor of shape (N, 3) representing the rotated locations.

    Notes
    -----
    - https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula
    """
    axis_norm = torch.norm(axis, dim=-1, keepdim=True)  # Ensure the axis is a unit vector
    axis = axis / torch.clamp(axis_norm, min=eps)

    cos_theta = torch.cos(angle).unsqueeze(-1)
    sin_theta = torch.sin(angle).unsqueeze(-1)

    s1 = v * cos_theta
    s2 = torch.cross(axis, v, dim=-1) * sin_theta
    s3 = axis * torch.sum(v * axis, dim=-1, keepdim=True) * (1 - cos_theta)
    v_rot = s1 + s2 + s3

    return v_rot


def compute_directions(source_coords: torch.Tensor, target_coords: torch.Tensor) -> torch.Tensor:
    """Compute the direction of the edge on the tangent plane.

    1. Rotate coordinate system so target_coords is at north pole (0, 0, 1)
    2. Rotate source_coords by the same rotation
    3. Project rotated source onto tangent plane at north pole (xy-plane)
    4. Direction = (x, y) components pointing toward rotated source

    Parameters
    ----------
    source_coords : torch.Tensor
        Source coordinates in lat/lon (radians), shape (N, 2)
    target_coords : torch.Tensor
        Target coordinates in lat/lon (radians), shape (N, 2)

    Returns
    -------
    torch.Tensor
        Unit direction vectors on tangent plane, shape (N, 2)
    """
    epsilon = 1e-8

    source_coords_xyz = latlon_rad_to_cartesian(source_coords, 1.0)
    target_coords_xyz = latlon_rad_to_cartesian(target_coords, 1.0)
    north_pole = torch.tensor([NORTH_POLE], dtype=source_coords.dtype, device=source_coords.device)

    # Compute dot product with north pole -> pole cases
    dot = (target_coords_xyz * north_pole).sum(dim=1, keepdim=True)

    # masks
    at_north = (dot > 1.0 - epsilon).squeeze(1)
    at_south = (dot < -1.0 + epsilon).squeeze(1)
    normal = ~(at_north | at_south)

    # init output
    direction = torch.zeros(
        (source_coords_xyz.shape[0], 2), dtype=source_coords_xyz.dtype, device=source_coords_xyz.device
    )

    # target at north pole - no rotation needed
    if at_north.any():
        xy = source_coords_xyz[at_north, :2]
        direction[at_north] = xy / torch.clamp(torch.linalg.norm(xy, dim=1, keepdim=True), min=epsilon)

    # target at south pole - rotate 180° around x-axis
    if at_south.any():
        rotated_south = source_coords_xyz[at_south].clone()
        # flip y and z
        rotated_south[:, 1] *= -1
        rotated_south[:, 2] *= -1
        xy = rotated_south[:, :2]
        direction[at_south] = xy / torch.clamp(torch.linalg.norm(xy, dim=1, keepdim=True), min=epsilon)

    # rotation via cross product
    if normal.any():
        target_normal = target_coords_xyz[normal]
        source_normal = source_coords_xyz[normal]

        axis = torch.cross(target_normal, north_pole.expand_as(target_normal), dim=1)
        sin_theta = torch.linalg.norm(axis, dim=1, keepdim=True)  # ||target x north_pole||
        cos_theta = torch.clamp((target_normal * north_pole).sum(dim=1, keepdim=True), -1.0, 1.0)
        theta = torch.atan2(sin_theta, cos_theta).squeeze(1)

        axis_unit = axis / torch.clamp(sin_theta, min=epsilon)

        rotated_normal = rotate_vectors(source_normal, axis_unit, theta)
        xy = rotated_normal[:, :2]
        direction[normal] = xy / torch.clamp(torch.linalg.norm(xy, dim=1, keepdim=True), min=epsilon)

    return direction



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import healpy as hp
import torch


def get_healpix_edgeindex(resolution: int) -> torch.Tensor:
    """Get the HEALPix edge index for a given resolution."""
    npix = hp.nside2npix(2**resolution)
    edge_index = torch.zeros((2, npix * 8))
    edge_index[0] = torch.repeat_interleave(torch.arange(npix), 8)
    for i in range(npix):
        edge_index[1, i * 8 : (i + 1) * 8] = torch.from_numpy(hp.get_all_neighbours(2**resolution, i, nest=True))
    return edge_index



# (C) Copyright 2024- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from abc import ABC
from abc import abstractmethod

import numpy as np
import torch
from scipy.spatial import ConvexHull
from scipy.spatial import SphericalVoronoi
from scipy.spatial import Voronoi
from torch_geometric.data.storage import NodeStorage

from anemoi.datasets import open_dataset
from anemoi.graphs import EARTH_RADIUS
from anemoi.graphs.generate.transforms import latlon_rad_to_cartesian_np
from anemoi.graphs.nodes.attributes.base_attributes import BaseNodeAttribute

LOGGER = logging.getLogger(__name__)


class BaseAreaWeights(BaseNodeAttribute, ABC):
    """Base class for area weights of the nodes."""

    def get_latlon_coordinates(self, nodes: NodeStorage) -> torch.Tensor:
        return nodes.x.to(torch.float64)

    @abstractmethod
    def compute_area_weights(self, latlons: np.ndarray) -> np.ndarray: ...

    def get_raw_values(self, nodes: NodeStorage, **kwargs) -> torch.Tensor:
        """Compute the weights.

        Parameters
        ----------
        nodes : NodeStorage
            Nodes of the graph.
        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        torch.Tensor
            Area weights for the nodes.
        """
        latlons = self.get_latlon_coordinates(nodes).cpu().numpy()
        area_weights = self.compute_area_weights(latlons)
        return torch.from_numpy(area_weights)


class UniformWeights(BaseAreaWeights):
    """Implements a uniform weight for the nodes.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the area attributes for each node.
    """

    def compute_area_weights(self, latlons: np.ndarray) -> np.ndarray:
        """Compute area weights.

        Parameters
        ----------
        latlons : np.ndarray
            2D array of shape (N, 2) with latitude and longitude coordinates of the
            nodes of the graph.

        Returns
        -------
        np.ndarray
            Ones.
        """
        return np.ones(latlons.shape[0])


class PlanarAreaWeights(BaseAreaWeights):
    """Planar area weights

    It computes the area in a 2D plane asociated to each node.

    Attributes
    ----------
    norm : str
        Normalisation of the weights.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the area attributes for each node.
    """

    def _compute_mean_nearest_distance(self, points: np.ndarray) -> float:
        """Compute mean distance to nearest neighbor for each point.

        Parameters
        ----------
        points : np.ndarray
            Array of point coordinates (N x 2)

        Returns
        -------
        float
            Mean nearest neighbor distance
        """
        from scipy.spatial import cKDTree

        tree = cKDTree(points)
        distances, _ = tree.query(points, k=2)
        return float(distances[:, 1].mean())

    def _get_boundary_ring(self, points: np.ndarray, resolution: float) -> np.ndarray:
        """Add a ring of boundary points around the input points.

        Parameters
        ----------
        points : np.ndarray
            Original point coordinates
        resolution : float
            Approximate spacing between points

        Returns
        -------
        np.ndarray
            Array including original and boundary points
        """
        # Get convex hull vertices
        hull = ConvexHull(points)
        hull_points = points[hull.vertices]

        # Expand hull outward
        centroid = np.mean(hull_points, axis=0)
        vectors = hull_points - centroid
        expanded_hull = hull_points + vectors * (2**0.5 * resolution / np.linalg.norm(vectors, axis=1)[:, np.newaxis])

        # Create points along each hull edge
        boundary_points = []
        p1 = expanded_hull
        p2 = np.roll(expanded_hull, 1, axis=0)

        # Calculate number of points needed along this edge
        edge_length = np.linalg.norm(p2 - p1, axis=1)
        num_points = np.ceil(edge_length / resolution).astype(int)

        for i in np.where(num_points > 2)[0]:
            # Create evenly spaced points along the edge
            t = np.linspace(0, 1, num_points[i])[1:-1][:, None]  # Exclude last point to avoid duplicates
            edge_points = p1[i] + t * (p2[i] - p1[i])
            boundary_points.append(edge_points)

        return np.concatenate([expanded_hull, np.vstack(boundary_points)])

    def compute_area_weights(self, latlons: np.ndarray) -> np.ndarray:
        """Compute area weights.

        Parameters
        ----------
        latlons : np.ndarray
            2D array of shape (N, 2) with latitude and longitude coordinates of the
            nodes of the graph.

        Returns
        -------
        np.ndarray
            Planar area weights.
        """
        resolution = self._compute_mean_nearest_distance(latlons)
        boundary_points = self._get_boundary_ring(latlons, resolution)

        # Compute convex hull over all points (boundary ring included)
        extended_points = np.vstack([latlons, boundary_points])
        v = Voronoi(extended_points, qhull_options="QJ Pp")

        # Compute the area of each node's region, excluding those in the boundary ring
        areas = []
        for idx in range(len(latlons)):
            p_idx = v.point_region[idx]
            r = v.regions[p_idx]
            poly_coords = v.vertices[r]
            area = ConvexHull(poly_coords).volume
            areas.append(area)

        return np.array(areas)


class MaskedPlanarAreaWeights(PlanarAreaWeights):
    """Masked planar area weights

    It computes the area in a 2D plane asociated to each node.

    Attributes
    ----------
    mask_node_attr_name : str
        Name of a node attribute to use as a mask for the computing the area weights.
        It sets to 0 values outside this masked region.
    norm : str, optional
        Normalisation of the weights.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the area attributes for each node.
    """

    def __init__(
        self,
        mask_node_attr_name: str,
        norm: str | None = None,
        dtype: str = "float32",
    ) -> None:
        super().__init__(norm, dtype)
        assert isinstance(
            mask_node_attr_name, str
        ), f"{self.__class__.__name__} requires a string for 'mask_node_attr_name' variable."
        self.mask_node_attr_name = mask_node_attr_name

    def get_raw_values(self, nodes: NodeStorage, **kwargs) -> torch.Tensor:
        assert self.mask_node_attr_name in nodes, f"Node attribute '{self.mask_node_attr_name}' not found in nodes."
        attr_values = super().get_raw_values(nodes, **kwargs).to(self.device)
        mask = nodes[self.mask_node_attr_name].squeeze()
        return attr_values * mask


class SphericalAreaWeights(BaseAreaWeights):
    """Spherical area weights

    It computes the area of a unit radius sphere asociated to each node.

    Attributes
    ----------
    norm : str
        Normalisation of the weights.
    radius : float
        Radius of the sphere.
    centre : np.ndarray
        Centre of the sphere.
    fill_value : float
        Value to fill the empty regions.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the area attributes for each node.
    """

    def __init__(
        self,
        norm: str | None = None,
        radius: float = 1.0,
        centre: np.ndarray = np.array([0, 0, 0]),
        fill_value: float = 0.0,
        dtype: str = "float32",
    ) -> None:
        assert isinstance(fill_value, float) or isinstance(
            fill_value, int
        ), f"fill_value must be float or nan but it is {type(fill_value)}"
        assert (
            isinstance(radius, float) or isinstance(radius, int)
        ) and radius > 0, f"radius must be a positive value, but radius={radius}"
        super().__init__(norm, dtype)
        self.radius = radius
        self.centre = centre
        self.fill_value = fill_value

    def compute_area_weights(self, latlons: np.ndarray) -> np.ndarray:
        """Compute the area associated to each node.

        It uses Voronoi diagrams to compute the area of each node.

        Parameters
        ----------
        latlons : np.ndarray
            2D array of shape (N, 2) with latitude and longitude coordinates of the
            nodes of the graph.

        Returns
        -------
        np.ndarray
            Spherical area weights.
        """
        points = latlon_rad_to_cartesian_np(latlons)
        sv = SphericalVoronoi(points, self.radius, self.centre)
        mask = np.array([bool(i) for i in sv.regions])
        sv.regions = [region for region in sv.regions if region]
        # compute the area weight without empty regions
        area_weights = sv.calculate_areas()
        if (null_nodes := (~mask).sum()) > 0:
            LOGGER.warning(
                "%s is filling %d (%.2f%%) nodes with value %f",
                self.__class__.__name__,
                null_nodes,
                100 * null_nodes / len(mask),
                self.fill_value,
            )
        result = np.ones(points.shape[0]) * self.fill_value
        result[mask] = area_weights
        LOGGER.debug(
            "There are %d of weights, which (unscaled) add up a total weight of %.2f.",
            len(result),
            result.sum(),
        )
        return result


class BaseLatWeightedAttribute(BaseAreaWeights, ABC):
    """Base class for latitude-weigthed area weights."""

    @abstractmethod
    def compute_latitude_weight(self, latitudes: np.ndarray) -> np.ndarray: ...

    def compute_area_weights(self, latlons: np.ndarray) -> np.ndarray:
        return self.compute_latitude_weight(latlons[:, 0])


class CosineLatWeightedAttribute(BaseLatWeightedAttribute):
    """Latitude-weighting of the node attributes for rectilinear grids.

    Attributes
    ----------
    min_value : float
        Minimum value of the weights when the latitude is -pi/2 or pi/2 radians.
    max_value : float
        Maximum value of the weights when the latitude is 0 radians.
    norm : str
        Normalisation of the weights.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the area attributes for each node.
    """

    def __init__(
        self,
        min_value: float = 1e-3,
        max_value: float = 1,
        norm: str | None = None,
        dtype: str = "float32",
    ) -> None:
        super().__init__(norm, dtype)
        self.min_value = min_value
        self.max_value = max_value

    def compute_latitude_weight(self, latitudes: np.ndarray) -> np.ndarray:
        return (self.max_value - self.min_value) * np.cos(latitudes) + self.min_value


class IsolatitudeAreaWeights(BaseLatWeightedAttribute):
    r"""Latitude-weighted area weights for rectilinear grids.

    Attributes
    ----------
    norm : str
        Normalisation of the weights.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the area attributes for each node.

    Notes
    ------
    The area of a latitude band is

    .. math::
        A = 2\pi R(\sin(lat_2) - \sin(lat_1))

    where R is the earth radius and lat_1, lat_2 are in radians.
    """

    def compute_latitude_weight(self, latitudes: np.ndarray) -> np.ndarray:
        # Get the latitudes defining the bands
        unique_lats = np.sort(np.unique(latitudes))
        divisory_lats = (unique_lats[1:] + unique_lats[:-1]) / 2
        divisory_lats = np.concatenate([[-np.pi / 2], divisory_lats, [np.pi / 2]])

        # Compute the latitude band area
        lat_1 = divisory_lats[1:]
        lat_2 = divisory_lats[:-1]
        assert np.all(lat_1 >= lat_2), "Nodes should be sorted by latitude."
        ring_area_km = 2 * np.pi * EARTH_RADIUS * (np.sin(lat_1) - np.sin(lat_2))

        # Compute the number of points/nodes at each latitude band
        lat_to_ring = {lat: idx for idx, lat in enumerate(unique_lats)}
        lat_rings = np.array([lat_to_ring[lat] for lat in latitudes])
        lat_counts = np.bincount(lat_rings, minlength=len(unique_lats))

        # Compute the area of each node
        area_km = dict(zip(unique_lats, ring_area_km / lat_counts))
        return np.array([area_km[lat] for lat in latitudes])


class AnemoiDatasetVariableWeights(BaseNodeAttribute):
    """Load area weights from a variable in the dataset.

    Attributes
    ----------
    variable : str
        Name of the variable to use as weights.
    norm : str, optional
        Method to use to normalise the weights.

    Methods
    -------
    get_raw_values(self, nodes)
        Extract the data and convert it to a Torch tensor object.
    """

    def __init__(self, variable: str, norm: str | None = None, dtype: str = "float32") -> None:
        super().__init__(norm, dtype)
        self.variable = variable

    def _read_data(self, nodes: NodeStorage, **kwargs) -> np.ndarray:
        """Read the weighting variable from the dataset."""
        return open_dataset(nodes["_dataset"], select=self.variable)[0].squeeze()

    def get_raw_values(self, nodes: NodeStorage, **kwargs) -> torch.Tensor:
        """Extract the data and convert it to a Torch tensor object.

        Parameters
        ----------
        nodes : NodeStorage
            Nodes of the graph.
        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        torch.Tensor
            Area weights for the nodes.
        """
        assert nodes["node_type"] in [
            "ZarrDatasetNodes",
            "AnemoiDatasetNodes",
        ], f"{self.__class__.__name__} can only be used with AnemoiDatasetNodes."
        data = torch.from_numpy(self._read_data(nodes))
        return self.post_process(data)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

import numpy as np
import torch
from sklearn.neighbors import NearestNeighbors
from torch_geometric.data.storage import NodeStorage
from torch_geometric.nn import knn

from anemoi.graphs.edges.builders.base import BaseDistanceEdgeBuilders

LOGGER = logging.getLogger(__name__)


class KNNEdges(BaseDistanceEdgeBuilders):
    """Computes KNN based edges and adds them to the graph.

    It uses as reference the target nodes.

    Attributes
    ----------
    source_name : str
        The name of the source nodes.
    target_name : str
        The name of the target nodes.
    num_nearest_neighbours : int
        Number of nearest neighbours.
    source_mask_attr_name : str | None
        The name of the source mask attribute to filter edge connections.
    target_mask_attr_name : str | None
        The name of the target mask attribute to filter edge connections.

    Methods
    -------
    register_edges(graph)
        Register the edges in the graph.
    register_attributes(graph, config)
        Register attributes in the edges of the graph.
    update_graph(graph, attrs_config)
        Update the graph with the edges.
    """

    def __init__(
        self,
        source_name: str,
        target_name: str,
        num_nearest_neighbours: int,
        source_mask_attr_name: str | None = None,
        target_mask_attr_name: str | None = None,
    ) -> None:
        super().__init__(source_name, target_name, source_mask_attr_name, target_mask_attr_name)
        assert isinstance(num_nearest_neighbours, int), "Number of nearest neighbours must be an integer."
        assert num_nearest_neighbours > 0, "Number of nearest neighbours must be positive."
        self.num_nearest_neighbours = num_nearest_neighbours

        LOGGER.info(
            "Using KNN-Edges (with %d nearest neighbours) between %s and %s.",
            self.num_nearest_neighbours,
            self.source_name,
            self.target_name,
        )

    def _compute_edge_index_pyg(self, source_coords: torch.Tensor, target_coords: torch.Tensor) -> torch.Tensor:
        edge_index = knn(source_coords, target_coords, k=self.num_nearest_neighbours)
        edge_index = torch.flip(edge_index, [0])
        return edge_index

    def _compute_adj_matrix_sklearn(self, source_coords: torch.Tensor, target_coords: torch.Tensor) -> np.ndarray:
        nearest_neighbour = NearestNeighbors(metric="euclidean", n_jobs=4)
        nearest_neighbour.fit(source_coords.cpu())
        adj_matrix = nearest_neighbour.kneighbors_graph(
            target_coords.cpu(),
            n_neighbors=self.num_nearest_neighbours,
        ).tocoo()

        return adj_matrix


class ReversedKNNEdges(KNNEdges):
    """Computes KNN based edges and adds them to the graph.

    It uses as reference the source nodes.

    Attributes
    ----------
    source_name : str
        The name of the source nodes.
    target_name : str
        The name of the target nodes.
    num_nearest_neighbours : int
        Number of nearest neighbours.
    source_mask_attr_name : str | None
        The name of the source mask attribute to filter edge connections.
    target_mask_attr_name : str | None
        The name of the target mask attribute to filter edge connections.

    Methods
    -------
    register_edges(graph)
        Register the edges in the graph.
    register_attributes(graph, config)
        Register attributes in the edges of the graph.
    update_graph(graph, attrs_config)
        Update the graph with the edges.
    """

    def get_cartesian_node_coordinates(
        self, source_nodes: NodeStorage, target_nodes: NodeStorage
    ) -> tuple[torch.Tensor, torch.Tensor]:
        source_coords, target_coords = super().get_cartesian_node_coordinates(source_nodes, target_nodes)
        return target_coords, source_coords

    def undo_masking_adj_matrix(self, adj_matrix, source_nodes: NodeStorage, target_nodes: NodeStorage):
        adj_matrix = adj_matrix.T
        return super().undo_masking_adj_matrix(adj_matrix, source_nodes, target_nodes)

    def undo_masking_edge_index(
        self, edge_index: torch.Tensor, source_nodes: NodeStorage, target_nodes: NodeStorage
    ) -> torch.Tensor:
        edge_index = torch.flip(edge_index, [0])
        return super().undo_masking_edge_index(edge_index, source_nodes, target_nodes)



import numpy as np
import torch

from anemoi.graphs.utils import concat_edges
from anemoi.graphs.utils import get_edge_attributes


def test_concat_edges():
    edge_indices1 = torch.tensor([[0, 1, 2, 3], [-1, -2, -3, -4]], dtype=torch.int64)
    edge_indices2 = torch.tensor(np.array([[0, 4], [-1, -5]]), dtype=torch.int64)
    no_edges = torch.tensor([[], []], dtype=torch.int64)

    result1 = concat_edges(edge_indices1, edge_indices2)
    result2 = concat_edges(no_edges, edge_indices2)

    expected1 = torch.tensor([[0, 1, 2, 3, 4], [-1, -2, -3, -4, -5]], dtype=torch.int64)

    assert torch.allclose(result1, expected1)
    assert torch.allclose(result2, edge_indices2)


def test_get_edge_attributes():
    mock_config = {
        "nodes": "mock_nodes_config",
        "edges": [
            {
                "source_name": "mock_nodes",
                "target_name": "mock_nodes",
                "attributes": {"attr": "attr_config"},
                "extra_keys": "extra_values",
            },
        ],
    }

    # test non-empty selection
    edge_attrs = get_edge_attributes(mock_config, "mock_nodes", "mock_nodes")
    expected = mock_config["edges"][0]["attributes"]
    assert edge_attrs == expected

    # test empty selection
    edge_attrs = get_edge_attributes(mock_config, "mock_nodes", "other_nodes")
    assert edge_attrs == {}

    edge_attrs = get_edge_attributes(mock_config, "other_nodes", "mock_nodes")
    assert edge_attrs == {}



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from collections.abc import Iterable

import networkx as nx
import numpy as np
import trimesh
from sklearn.neighbors import BallTree

from anemoi.graphs.generate.masks import KNNAreaMaskBuilder
from anemoi.graphs.generate.transforms import cartesian_to_latlon_rad
from anemoi.graphs.generate.utils import get_coordinates_ordering

LOGGER = logging.getLogger(__name__)


def create_tri_nodes(
    resolution: int, area_mask_builder: KNNAreaMaskBuilder | None = None
) -> tuple[nx.DiGraph, np.ndarray, list[int]]:
    """Creates a global mesh from a refined icosahedron.

    This method relies on the trimesh python library.

    Parameters
    ----------
    resolution : int
        Level of mesh resolution to consider.
    area_mask_builder : KNNAreaMaskBuilder
        KNNAreaMaskBuilder with the cloud of points to limit the mesh area, by default None.

    Returns
    -------
    graph : networkx.Graph
        The specified graph (only nodes) sorted by latitude and longitude.
    coords_rad : np.ndarray
        The node coordinates (not ordered) in radians.
    node_ordering : list[int]
        Order of the node coordinates to be sorted by latitude and longitude.
    """
    coords_rad = get_latlon_coords_icosphere(resolution)

    node_ordering = get_coordinates_ordering(coords_rad)

    if area_mask_builder is not None:
        area_mask = area_mask_builder.get_mask(coords_rad)
        node_ordering = node_ordering[area_mask[node_ordering]]

    # Creates the graph, with the nodes sorted by latitude and longitude.
    nx_graph = create_nx_graph_from_tri_coords(coords_rad, node_ordering)

    return nx_graph, coords_rad, list(node_ordering)


def create_stretched_tri_nodes(
    base_resolution: int,
    lam_resolution: int,
    area_mask_builder: KNNAreaMaskBuilder | None = None,
) -> tuple[nx.DiGraph, np.ndarray, list[int]]:
    """Creates a global mesh with 2 levels of resolution.

    The base resolution is used to define the nodes outside the Area Of Interest (AOI),
    while the lam_resolution is used to define the nodes inside the AOI.

    Parameters
    ---------
    base_resolution : int
        Global resolution level.
    lam_resolution : int
        Local resolution level.
    area_mask_builder : KNNAreaMaskBuilder
        NearestNeighbors with the cloud of points to limit the mesh area.

    Returns
    -------
    nx_graph : nx.DiGraph
        The graph with the added nodes.
    coords_rad : np.ndarray
        The node coordinates (not ordered) in radians.
    node_ordering : list[int]
        Order of the node coordinates to be sorted by latitude and longitude.
    """
    assert area_mask_builder is not None, "AOI mask builder must be provided to build refined grid."
    # Get the low resolution nodes
    base_coords_rad = get_latlon_coords_icosphere(base_resolution)

    # Define the low resolution outside AOI mask
    base_area_mask = ~area_mask_builder.get_mask(base_coords_rad)

    # Get the high resolution nodes
    coords_rad = get_latlon_coords_icosphere(lam_resolution)

    # Get the node ordering for all high resolution nodes
    node_ordering = get_coordinates_ordering(coords_rad)

    # Define the high resolution inside AOI mask
    lam_area_mask = area_mask_builder.get_mask(coords_rad)

    # Pad the low resolution ~(AOI mask) to match the high resolution AOI mask
    base_area_mask_padded = np.pad(base_area_mask, (0, len(lam_area_mask) - len(base_area_mask)), mode="constant")

    # Define the final mask (low resolution outside AOI | high resolution inside AOI )
    area_mask = np.logical_or(base_area_mask_padded, lam_area_mask)

    # Redefine the node ordering to include final node selection
    node_ordering = node_ordering[area_mask[node_ordering]]

    # Creates the graph, with the nodes sorted by latitude and longitude.
    nx_graph = create_nx_graph_from_tri_coords(coords_rad, node_ordering)

    return nx_graph, coords_rad, list(node_ordering)


def get_latlon_coords_icosphere(resolution: int) -> np.ndarray:
    """Get the latitude and longitude coordinates (in radians) of the icosphere.

    Parameters
    ----------
    resolution : int
        The resolution level of the icosphere.

    Returns
    -------
    np.ndarray
        The latitude and longitude coordinates, in radians, of the icosphere.
    """
    sphere = trimesh.creation.icosphere(subdivisions=resolution, radius=1.0)
    coords_rad = cartesian_to_latlon_rad(sphere.vertices)
    return coords_rad


def create_nx_graph_from_tri_coords(coords_rad: np.ndarray, node_ordering: np.ndarray) -> nx.DiGraph:
    """Creates the networkx graph from the coordinates and the node ordering."""
    graph = nx.DiGraph()
    for i, coords in enumerate(coords_rad[node_ordering]):
        node_id = node_ordering[i]
        graph.add_node(node_id, hcoords_rad=coords)

    assert list(graph.nodes.keys()) == list(node_ordering), "Nodes are not correctly added to the graph."
    assert graph.number_of_nodes() == len(node_ordering), "The number of nodes must be the same."
    return graph


def add_1_hop_edges(
    nodes_coords_rad,
    node_resolutions: list[int],
    edge_resolutions: list[int],
    node_ordering: list[int],
    area_mask_builder: KNNAreaMaskBuilder | None = None,
) -> np.ndarray:
    """Adds edges for x_hops = 1 relying on trimesh only."""

    hop_1_edges = []

    # Loop over the edge_resolutions to get edges at all refinement levels
    for subdivisions in edge_resolutions:
        sphere = trimesh.creation.icosphere(subdivisions=subdivisions, radius=1.0)
        LOGGER.debug("Adding %d unmasked 1-hop edges for resolution %d", sphere.edges.shape[0], subdivisions)
        hop_1_edges.append(sphere.edges)

    # Concatenate all edges from different resolutions and transpose to get shape (2, num_edges)
    multiscale_edges = np.transpose(np.concatenate(hop_1_edges, axis=0), (1, 0))

    # Map the edges to the node ordering
    #  - Calculate the total (unmasked) number of nodes
    unmasked_nodes = cartesian_to_latlon_rad(
        trimesh.creation.icosphere(subdivisions=max(node_resolutions), radius=1.0).vertices
    )
    assert np.all(
        (nodes_coords_rad.cpu().numpy() - unmasked_nodes[node_ordering]) == 0
    ), "Node coordates do not match coordinates used for multi-scale edge building"
    LOGGER.debug("unmasked_nodes shape[0]: %d", unmasked_nodes.shape[0])
    if area_mask_builder is not None:
        # Take care of the edges with start- or end-point outside the mask
        inverse_ordering = np.full(unmasked_nodes.shape[0], -1, dtype=int)
    else:
        inverse_ordering = np.full(unmasked_nodes.shape[0], 0, dtype=int)

    # Update the start- and end indexes according to the node ordering
    inverse_ordering[node_ordering] = np.arange(len(node_ordering))
    updated_edges = inverse_ordering[multiscale_edges]
    valid_edges_mask = np.all(updated_edges >= 0, axis=0)

    # Select only those edges requested by the mask
    multiscale_edges = updated_edges[:, valid_edges_mask]
    LOGGER.debug("multiscale_edges_shape: %s", multiscale_edges.shape)

    return multiscale_edges


def add_edges_to_nx_graph(
    graph: nx.DiGraph,
    resolutions: list[int],
    x_hops: int = 1,
    area_mask_builder: KNNAreaMaskBuilder | None = None,
) -> nx.DiGraph:
    """Adds the edges to the graph.

    This method adds multi-scale connections to the existing graph. The corresponfing nodes or vertices
    are defined by an isophere at the different esolutions (or refinement levels) specified.

    Parameters
    ----------
    graph : nx.DiGraph
        The graph to add the edges. It should correspond to the mesh nodes, without edges.
    resolutions : list[int]
        Levels of mesh refinement levels to consider.
    x_hops : int, optional
        Number of hops between 2 nodes to consider them neighbours, by default 1.
    area_mask_builder : KNNAreaMaskBuilder
        NearestNeighbors with the cloud of points to limit the mesh area, by default None.

    Returns
    -------
    graph : nx.DiGraph
        The graph with the added edges.
    """
    assert x_hops > 0, "x_hops == 0, graph would have no edges ..."

    graph_vertices = np.array([graph.nodes[i]["hcoords_rad"] for i in sorted(graph.nodes)])
    tree = BallTree(graph_vertices, metric="haversine")

    # Build the multi-scale connections
    for resolution in resolutions:
        # Define the coordinates of the isophere vertices at specified 'resolution' level
        r_sphere = trimesh.creation.icosphere(subdivisions=resolution, radius=1.0)
        r_vertices_rad = cartesian_to_latlon_rad(r_sphere.vertices)

        # Limit area of mesh points.
        if area_mask_builder is not None:
            area_mask = area_mask_builder.get_mask(r_vertices_rad)
            valid_nodes = np.where(area_mask)[0]
        else:
            valid_nodes = None

        node_neighbours = get_neighbours_within_hops(r_sphere, x_hops, valid_nodes=valid_nodes)

        _, vertex_mapping_index = tree.query(r_vertices_rad, k=1)
        neighbour_pairs = create_node_neighbours_list(graph, node_neighbours, vertex_mapping_index)
        LOGGER.debug("Adding %d edges for resolution %d", len(neighbour_pairs), resolution)
        graph.add_edges_from(neighbour_pairs)
    return graph


def get_neighbours_within_hops(
    tri_mesh: trimesh.Trimesh, x_hops: int, valid_nodes: list[int] | None = None
) -> dict[int, set[int]]:
    """Get the neigbour connections in the graph.

    Parameters
    ----------
    tri_mesh : trimesh.Trimesh
        The mesh to consider.
    x_hops : int
        Number of hops between 2 nodes to consider them neighbours.
    valid_nodes : list[int], optional
        List of valid nodes to consider, by default None. It is useful to consider only a subset of the nodes to save
        computation time.

    Returns
    -------
    neighbours : dict[int, set[int]]
        A list with the neighbours for each vertex. The element at position 'i' correspond to the neighbours to the
        i-th vertex of the mesh.
    """
    edges = tri_mesh.edges_unique

    if valid_nodes is not None:
        edges = edges[np.isin(tri_mesh.edges_unique, valid_nodes).all(axis=1)]
    else:
        valid_nodes = list(range(len(tri_mesh.vertices)))
    graph = nx.from_edgelist(edges)

    neighbours = {
        i: set(nx.ego_graph(graph, i, radius=x_hops, center=False) if i in graph else []) for i in valid_nodes
    }

    return neighbours


def add_neigbours_edges(
    graph: nx.Graph,
    node_idx: int,
    neighbour_indices: Iterable[int],
    self_loops: bool = False,
    vertex_mapping_index: np.ndarray | None = None,
) -> nx.Graph:
    """Adds the edges of one node to its neighbours.

    Parameters
    ----------
    graph : nx.Graph
        The graph.
    node_idx : int
        The node considered.
    neighbour_indices : list[int]
        The neighbours of the node.
    self_loops : bool, optional
        Whether is supported to add self-loops, by default False.
    vertex_mapping_index : np.ndarray, optional
        Index to map the vertices from the refined sphere to the original one, by default None.

    Returns
    -------
    nx.Graph
        The graph with the added edges.
    """
    graph_nodes_idx = list(sorted(graph.nodes))
    for neighbour_idx in neighbour_indices:
        if not self_loops and node_idx == neighbour_idx:  # no self-loops
            continue

        if vertex_mapping_index is not None:
            # Use the same method to add edge in all spheres
            node_neighbour = graph_nodes_idx[vertex_mapping_index[neighbour_idx][0]]
            node = graph_nodes_idx[vertex_mapping_index[node_idx][0]]
        else:
            node_neighbour = graph_nodes_idx[neighbour_idx]
            node = graph_nodes_idx[node_idx]

        # add edge to the graph
        if node in graph and node_neighbour in graph:
            graph.add_edge(node_neighbour, node)

    return graph


def create_node_neighbours_list(
    graph: nx.Graph,
    node_neighbours: dict[int, set[int]],
    vertex_mapping_index: np.ndarray | None = None,
    self_loops: bool = False,
) -> list[tuple]:
    """Preprocesses the dict of node neighbours.

    Parameters:
    -----------
    graph: nx.Graph
        The graph.
    node_neighbours: dict[int, set[int]]
        dictionairy with key: node index and value: set of neighbour node indices
    vertex_mapping_index: np.ndarry
        Index to map the vertices from the refined sphere to the original one, by default None.
    self_loops: bool
        Whether is supported to add self-loops, by default False.

    Returns:
    --------
    list: tuple
        A list with containing node neighbour pairs in tuples
    """
    graph_nodes_idx = list(sorted(graph.nodes))

    if vertex_mapping_index is None:
        vertex_mapping_index = np.arange(len(graph.nodes)).reshape(len(graph.nodes), 1)

    neighbour_pairs = [
        (graph_nodes_idx[vertex_mapping_index[node_neighbour][0]], graph_nodes_idx[vertex_mapping_index[node][0]])
        for node, neighbours in node_neighbours.items()
        for node_neighbour in neighbours
        if node != node_neighbour or (self_loops and node == node_neighbour)
    ]

    return neighbour_pairs



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

import torch

LOGGER = logging.getLogger(__name__)


class NormaliserMixin:
    """Mixin class for normalising values.

    This class provides a method to normalise values within each group specified by an index tensor.
    Supported normalisation methods: None, 'l1', 'l2', 'unit-max', 'unit-range', 'unit-std'.
    """

    def compute_nongrouped_statistics(self, values: torch.Tensor, *_args) -> tuple[float, ...]:
        if self.norm == "l1":
            statistics = (torch.sum(values),)

        elif self.norm == "l2":
            statistics = (torch.norm(values),)

        elif self.norm == "unit-max":
            statistics = (torch.amax(values),)

        elif self.norm == "unit-range":
            statistics = torch.amin(values), torch.amax(values)

        elif self.norm == "unit-std":
            std = torch.std(values)
            if std == 0:
                LOGGER.warning(f"Std. dev. of the {self.__class__.__name__} values is 0. Normalisation is skipped.")
                return (1,)

            statistics = (std,)

        assert (
            statistics[-1] != 0
        ), f"Normalisation by zero encountered in {self.__class__.__name__} with norm '{self.norm}'."
        return statistics

    def compute_grouped_statistics(
        self, values: torch.Tensor, index: torch.Tensor, num_groups: int, dtype, device
    ) -> tuple[torch.Tensor, ...]:
        if self.norm == "l1":
            group_sum = torch.zeros(num_groups, values.shape[1], dtype=dtype, device=device)
            group_sum = group_sum.index_add(0, index, values)
            group_statistics = (group_sum[index],)

        elif self.norm == "l2":
            group_sq = torch.zeros(num_groups, values.shape[1], dtype=dtype, device=device)
            group_sq = group_sq.index_add(0, index, values**2)
            group_norm = torch.sqrt(group_sq)
            group_statistics = (group_norm[index],)

        elif self.norm == "unit-max":
            group_max = torch.full((num_groups, values.shape[1]), float("-inf"), dtype=dtype, device=device)
            group_max = group_max.index_reduce(0, index, values, reduce="amax")
            group_statistics = (group_max[index],)

        elif self.norm == "unit-range":
            group_min = torch.full((num_groups, values.shape[1]), float("inf"), dtype=dtype, device=device)
            group_max = torch.full((num_groups, values.shape[1]), float("-inf"), dtype=dtype, device=device)
            group_min = group_min.index_reduce(0, index, values, reduce="amin")
            group_max = group_max.index_reduce(0, index, values, reduce="amax")
            denom = group_max - group_min
            denom[denom == 0] = 1  # avoid division by zero
            group_statistics = group_min[index], denom[index]

        elif self.norm == "unit-std":
            # Compute mean
            group_sum = torch.zeros(num_groups, values.shape[1], dtype=dtype, device=device)
            group_count = torch.zeros(num_groups, values.shape[1], dtype=dtype, device=device)
            ones = torch.ones_like(values)
            group_sum = group_sum.index_add(0, index, values)
            group_count = group_count.index_add(0, index, ones)
            group_mean = group_sum / group_count.clamp(min=1)
            # Compute variance
            mean_expanded = group_mean[index]
            sq_diff = (values - mean_expanded) ** 2
            group_var = torch.zeros_like(group_sum)
            group_var = group_var.index_add(0, index, sq_diff)
            group_var = group_var / group_count.clamp(min=1)
            group_std = torch.sqrt(group_var)
            # Avoid division by zero
            group_std[group_std == 0] = 1
            group_statistics = (group_std[index],)

        assert torch.all(
            group_statistics[-1] != 0
        ), f"Normalisation by zero encountered in {self.__class__.__name__} with norm '{self.norm}'."
        return group_statistics

    def normalise(self, values: torch.Tensor, *args) -> torch.Tensor:
        """Normalise the given values.

        It supports different normalisation methods: None, 'l1',
        'l2', 'unit-max', 'unit-range' and 'unit-std'.

        Parameters
        ----------
        values : torch.Tensor of shape (N, M)
            Values to normalise.

        Returns
        -------
        torch.Tensor
            Normalised values.
        """
        assert self.norm in {
            None,
            "l1",
            "l2",
            "unit-std",
            "unit-max",
            "unit-range",
            "log1p",
        }, f"Attribute normalisation '{self.norm}' is not valid. Options are: 'l1', 'l2', 'unit-max', 'unit-range', 'unit-std', 'log1p'."

        if self.norm == "log1p":
            return torch.log1p(values)

        if self.norm is None:
            LOGGER.debug(f"{self.__class__.__name__} values are not normalised.")
            return values

        if self.norm_by_group:
            statistics = self.compute_grouped_statistics(values, *args, device=values.device, dtype=values.dtype)
        else:
            statistics = self.compute_nongrouped_statistics(values, *args)

        if self.norm == "unit-range":
            values = values - statistics[0]

        return values / statistics[-1]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import numpy as np
import scipy
from typeguard import typechecked


def get_coordinates_ordering(coords: np.ndarray) -> np.ndarray:
    """Sort node coordinates by latitude and longitude.

    Parameters
    ----------
    coords : np.ndarray of shape (N, 2)
        The node coordinates, with the latitude in the first column and the
        longitude in the second column.

    Returns
    -------
    np.ndarray
        The order of the node coordinates to be sorted by latitude and longitude.
    """
    # Get indices to sort points by lon & lat in radians.
    index_latitude = np.argsort(coords[:, 1])
    index_longitude = np.argsort(coords[index_latitude][:, 0])[::-1]
    node_ordering = np.arange(coords.shape[0])[index_latitude][index_longitude]
    return node_ordering


@typechecked
def convert_list_to_adjacency_matrix(list_matrix: np.ndarray, ncols: int = 0) -> scipy.sparse.csr_matrix:
    """Convert an edge list into an adjacency matrix.

    Parameters
    ----------
    list_matrix : np.ndarray
        boolean matrix given by list of column indices for each row.
    ncols : int
        number of columns in result matrix.

    Returns
    -------
    scipy.sparse.csr_matrix
        sparse matrix [nrows, ncols]
    """
    nrows, ncols_per_row = list_matrix.shape
    indptr = np.arange(ncols_per_row * (nrows + 1), step=ncols_per_row)
    indices = list_matrix.ravel()
    return scipy.sparse.csr_matrix((np.ones(nrows * ncols_per_row), indices, indptr), dtype=bool, shape=(nrows, ncols))


@typechecked
def convert_adjacency_matrix_to_list(
    adj_matrix: scipy.sparse.csr_matrix,
    ncols_per_row: int,
    remove_duplicates: bool = True,
) -> np.ndarray:
    """Convert an adjacency matrix into an edge list.

    Parameters
    ----------
    adj_matrix : scipy.sparse.csr_matrix
        sparse (boolean) adjacency matrix
    ncols_per_row : int
        number of nonzero entries per row
    remove_duplicates : bool
        logical flag: remove duplicate rows.

    Returns
    -------
    np.ndarray
        boolean matrix given by list of column indices for each row.
    """
    if remove_duplicates:
        # The edges-vertex adjacency matrix may have duplicate rows, remove
        # them by selecting the rows that are unique:
        nrows = int(adj_matrix.nnz // ncols_per_row)
        mat = adj_matrix.indices.reshape((nrows, ncols_per_row))
        return np.unique(mat, axis=0)

    nrows = adj_matrix.shape[0]
    return adj_matrix.indices.reshape((nrows, ncols_per_row))


@typechecked
def selection_matrix(idx: np.ndarray, num_diagonals: int) -> scipy.sparse.csr_matrix:
    """Create a diagonal selection matrix.

    Parameters
    ----------
    idx : np.ndarray
        integer array of indices
    num_diagonals : int
        size of (square) selection matrix

    Returns
    -------
    scipy.sparse.csr_matrix
        diagonal matrix with ones at selected indices (idx,idx).
    """
    return scipy.sparse.csr_matrix((np.ones(len(idx)), (idx, idx)), dtype=bool, shape=(num_diagonals, num_diagonals))



# (C) Copyright 2024- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from abc import ABC
from abc import abstractmethod

import torch
from torch_geometric.data import HeteroData
from torch_geometric.data.storage import NodeStorage

from anemoi.graphs.normalise import NormaliserMixin
from anemoi.graphs.utils import get_distributed_device

LOGGER = logging.getLogger(__name__)


class BaseNodeAttribute(ABC, NormaliserMixin):
    """Base class for the weights of the nodes."""

    norm_by_group: bool = False

    def __init__(self, norm: str | None = None, dtype: str = "float32") -> None:
        self.norm = norm
        self.dtype = getattr(torch, dtype)
        self.device = get_distributed_device()

    @abstractmethod
    def get_raw_values(self, nodes: NodeStorage, **kwargs) -> torch.Tensor: ...

    def post_process(self, values: torch.Tensor) -> torch.Tensor:
        """Post-process the values."""
        if values.ndim == 1:
            values = torch.unsqueeze(values, -1)

        return self.normalise(values)

    def compute(self, graph: HeteroData, nodes_name: str, **kwargs) -> torch.Tensor:
        """Get the nodes attribute.

        Parameters
        ----------
        graph : HeteroData
            Graph.
        nodes_name : str
            Name of the nodes.
        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        torch.Tensor
            Attributes associated to the nodes.
        """
        assert (
            nodes_name in graph.node_types
        ), f"{nodes_name} is not a valid nodes name. The current graph has the following nodes: {graph.node_types}"
        nodes = graph[nodes_name].to(self.device)
        attributes = self.get_raw_values(nodes, **kwargs).to(dtype=self.dtype, device=self.device)
        return self.post_process(attributes)


class BooleanBaseNodeAttribute(BaseNodeAttribute, ABC):
    """Base class for boolean node attributes."""

    def __init__(self) -> None:
        super().__init__(norm=None, dtype="bool")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.nodes.attributes import SphericalAreaWeights
from anemoi.graphs.nodes.attributes import UniformWeights
from anemoi.graphs.nodes.builders.from_file import NPZFileNodes


@pytest.mark.parametrize("resolution", ["o16", "o48", "5km5"])
def test_init(mock_grids_path: tuple[str, int], resolution: str):
    """Test NPZFileNodes initialization."""
    grid_definition_path, _ = mock_grids_path
    node_builder = NPZFileNodes(
        npz_file=grid_definition_path + f"/grid-{resolution}.npz",
        name="test_nodes",
        lat_key="latitudes",
        lon_key="longitudes",
    )
    assert isinstance(node_builder, NPZFileNodes)


def test_fail_init_wrong_path():
    """Test NPZFileNodes initialization with invalid path."""
    node_builder = NPZFileNodes("invalid_path.myext", "test_nodes")
    with pytest.raises(AssertionError):
        node_builder.get_coordinates()


@pytest.mark.parametrize("resolution", ["o16", "o48", "5km5"])
def test_register_nodes(mock_grids_path: str, resolution: str):
    """Test NPZFileNodes register correctly the nodes."""
    graph = HeteroData()
    grid_definition_path, num_nodes = mock_grids_path
    node_builder = NPZFileNodes(npz_file=grid_definition_path + f"/grid-{resolution}.npz", name="test_nodes")

    graph = node_builder.register_nodes(graph)

    assert graph["test_nodes"].x is not None
    assert isinstance(graph["test_nodes"].x, torch.Tensor)
    assert graph["test_nodes"].x.shape == (num_nodes, 2)
    assert graph["test_nodes"].node_type == "NPZFileNodes"


@pytest.mark.parametrize("attr_class", [UniformWeights, SphericalAreaWeights])
@pytest.mark.parametrize("resolution", ["o16", "o48", "5km5"])
def test_register_attributes(
    graph_with_nodes: HeteroData, mock_grids_path: tuple[str, int], attr_class, resolution: str
):
    """Test NPZFileNodes register correctly the weights."""
    grid_definition_path, _ = mock_grids_path
    node_builder = NPZFileNodes(npz_file=grid_definition_path + f"/grid-{resolution}.npz", name="test_nodes")
    config = {"test_attr": {"_target_": f"anemoi.graphs.nodes.attributes.{attr_class.__name__}"}}

    graph = node_builder.register_attributes(graph_with_nodes, config)

    assert graph["test_nodes"]["test_attr"] is not None
    assert isinstance(graph["test_nodes"]["test_attr"], torch.Tensor)
    assert graph["test_nodes"]["test_attr"].shape[0] == graph["test_nodes"].x.shape[0]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import argparse

from anemoi.graphs.describe import GraphDescriptor
from anemoi.graphs.inspect import GraphInspector

from . import Command


def parse_str_area(area_str: str) -> tuple[float, float, float, float]:
    # Remove parentheses if present, then split by comma
    area_str = area_str.strip("()")
    parts = area_str.split(",")
    if len(parts) != 4:
        raise argparse.ArgumentTypeError("Area must be four comma-separated values: north,west,south,east")
    try:
        return [float(x) for x in parts]
    except ValueError:
        raise argparse.ArgumentTypeError("All area values must be floats.")


class Inspect(Command):
    """Inspect a graph.

    Example
    -------
    To save all visualizations in a folder, use the following command:
    ```
    anemoi-graphs inspect graph.pt output_dir/
    ```
    For high-resolution graphs, you can specify an area to crop the graph using the `--area` option:
    ```
    anemoi-graphs inspect graph.pt output_dir/ --area=40,10,20,30
    ```
    """

    internal = True
    timestamp = True

    def add_arguments(self, command_parser):
        command_parser.add_argument(
            "--show_attribute_distributions",
            action=argparse.BooleanOptionalAction,
            default=True,
            help="Hide distribution plots of edge/node attributes.",
        )
        command_parser.add_argument(
            "--area",
            type=parse_str_area,
            default=None,
            help="Area of interest to crop the nodes, (north, west, south, east).",
        )
        command_parser.add_argument(
            "--show_nodes",
            action=argparse.BooleanOptionalAction,
            default=False,
            help="Show the nodes of the graph.",
        )
        command_parser.add_argument(
            "--description",
            action=argparse.BooleanOptionalAction,
            default=True,
            help="Hide the description of the graph.",
        )
        command_parser.add_argument("path", help="Path to the graph (a .PT file).")
        command_parser.add_argument("output_path", help="Path to store the inspection results.")

    def run(self, args):
        kwargs = vars(args)

        if kwargs.get("description", False):
            GraphDescriptor(kwargs["path"]).describe()

        inspector = GraphInspector(**kwargs)
        inspector.inspect()


command = Inspect



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

import numpy as np
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.generate.masks import KNNAreaMaskBuilder
from anemoi.graphs.nodes.builders.base import BaseNodeBuilder

LOGGER = logging.getLogger(__name__)


class HEALPixNodes(BaseNodeBuilder):
    """Nodes from HEALPix grid.

    HEALPix is an acronym for Hierarchical Equal Area isoLatitude Pixelization of a sphere.

    Attributes
    ----------
    resolution : int
        The resolution of the grid.

    Methods
    -------
    get_coordinates()
        Get the lat-lon coordinates of the nodes.
    register_nodes(graph, name)
        Register the nodes in the graph.
    register_attributes(graph, name, config)
        Register the attributes in the nodes of the graph specified.
    update_graph(graph, name, attrs_config)
        Update the graph with new nodes and attributes.
    """

    def __init__(self, resolution: int, name: str) -> None:
        """Initialize the HEALPixNodes builder."""
        self.resolution = resolution
        super().__init__(name)
        self.hidden_attributes = BaseNodeBuilder.hidden_attributes | {"resolution"}

        assert isinstance(resolution, int), "Resolution must be an integer."
        assert resolution > 0, "Resolution must be positive."

    def get_coordinates(self) -> torch.Tensor:
        """Get the coordinates of the nodes.

        Returns
        -------
        torch.Tensor of shape (num_nodes, 2)
            Coordinates of the nodes, in radians.
        """
        import healpy as hp

        spatial_res_degrees = hp.nside2resol(2**self.resolution, arcmin=True) / 60
        LOGGER.info(f"Creating HEALPix nodes with resolution {spatial_res_degrees:.2} deg.")

        npix = hp.nside2npix(2**self.resolution)
        hpxlon, hpxlat = hp.pix2ang(2**self.resolution, range(npix), nest=True, lonlat=True)

        return self.reshape_coords(hpxlat, hpxlon)


class LimitedAreaHEALPixNodes(HEALPixNodes):
    """Nodes from HEALPix grid using an area of interest."""

    def __init__(
        self,
        resolution: str,
        reference_node_name: str,
        name: str,
        mask_attr_name: str | None = None,
        margin_radius_km: float = 100.0,
    ) -> None:
        super().__init__(resolution, name)
        self.area_mask_builder = KNNAreaMaskBuilder(reference_node_name, margin_radius_km, mask_attr_name)

    def register_nodes(self, graph: HeteroData) -> None:
        self.area_mask_builder.fit(graph)
        return super().register_nodes(graph)

    def get_coordinates(self) -> np.ndarray:
        coords = super().get_coordinates()

        LOGGER.info(
            'Limiting the "%s" nodes to a radius of %.2f km from the nodes of interest.',
            self.name,
            self.area_mask_builder.margin_radius_km,
        )
        area_mask = self.area_mask_builder.get_mask(coords)

        LOGGER.info('Masking out %d nodes from "%s".', len(area_mask) - area_mask.sum(), self.name)
        coords = coords[area_mask]

        return coords



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from pathlib import Path

import numpy as np
import torch
from jinja2 import Template
from torch_geometric.data import HeteroData

HTML_TEMPLATE_PATH = Path(__file__).parent / "interactive_3d.html.jinja"


def extract_nodes_edges(
    graph: HeteroData,
    nodes: list[str] | None = None,
    edges: list[str] | None = None,
) -> tuple[dict[str, torch.Tensor], dict[str, torch.Tensor]]:
    """Extracts nodes and edges from a heterogeneous graph.

    Optionally filters by specified types.

    Parameters
    ----------
    graph : HeteroData
        The graph to subset.
    nodes : list[str]
        List of node types to extract.
    edges : list[str]
        List of edge types to extract.

    Returns
    -------
    tuple[dict[str, torch.Tensor], dict[str, torch.Tensor]]
        A tuple containing two dictionaries:
        - The first dictionary maps node types to their coordinates tensors.
        - The second dictionary maps edge types to their edge index tensors.
    """
    if nodes:
        for n in nodes:
            if n not in graph.node_types:
                raise ValueError(f"Node type '{n}' not found in the graph.")
    else:
        nodes = graph.node_types

    if edges:
        edges_keys = ()
        for e in edges:
            edge_key = e.split("_to_")
            edge_key = (edge_key[0], "to", edge_key[1])
            if edge_key not in graph.edge_types:
                raise ValueError(f"Edge type '{e}' not found in the graph.")
            edges_keys += (edge_key,)
    else:
        edges_keys = graph.edge_types

    out_nodes = {n: graph[n].x for n in nodes}
    out_edges = {"_".join(e): graph[e].edge_index for e in edges_keys}

    return out_nodes, out_edges


def coords_to_latlon(coordinates):
    """Convert radians coordinates to latitude and longitude in degrees."""
    coordinates = np.rad2deg(coordinates)
    lats, lons = coordinates.T
    return lats, lons


def to_nodes_json(lats, lons, prefix="P"):
    """Convert nodes dictionary to JSON format for HTML rendering."""
    assert len(lats) == len(lons)
    names = [f"{prefix}_{i}" for i in range(len(lats))]
    points = [{"name": n, "pos": [float(lat), float(lon)]} for n, lat, lon in zip(names, lats, lons)]
    return points


def to_edges_json(names1, names2, pairs):
    edges = [[names1[i], names2[j]] for i, j in pairs]
    return edges


def plot_interactive_graph_3d(
    graph: HeteroData,
    out_file: str | Path,
) -> None:
    """Plot the entire graph in 3D.

    This method creates an interactive 3D visualization of the entire graph.

    Parameters
    ----------
    graph : dict
        The graph to plot.
    out_file : str | Path, optional
        Name of the file to save the plot. Default is None.
    """
    nodes, edges = extract_nodes_edges(graph)

    for node_set in nodes:
        node_lats, node_lons = coords_to_latlon(nodes[node_set].numpy())
        nodes[node_set] = to_nodes_json(node_lats, node_lons, prefix=node_set)

    for edge_set in edges:
        src_nodes, dst_nodes = edge_set.split("_to_")
        src_names = [f"{src_nodes}_{i}" for i in range(len(nodes[src_nodes]))]
        dst_names = [f"{dst_nodes}_{i}" for i in range(len(nodes[dst_nodes]))]
        edges[edge_set] = to_edges_json(src_names, dst_names, edges[edge_set].numpy().T)

    colors = [
        "#5050ff",
        "#ff5050",
        "#50ff50",
        "#ffaa00",
        "#aa00ff",
        "#00aaff",
        "#ff0055",
        "#55ff00",
        "#00ffaa",
        "#ff5500",
        "#0055ff",
        "#aa5500",
    ]

    nodes_embed = []
    for i, (node_set, pts) in enumerate(nodes.items()):
        nodes_embed.append({"name": node_set, "points": pts, "color": colors[i % len(colors)], "radius": 20.1 + i * 2})

    edges_embed = []
    for i, (edge_set, eds) in enumerate(edges.items()):
        edges_embed.append({"name": edge_set, "edges": eds})

    # # Render and save
    with open(HTML_TEMPLATE_PATH, "r") as f:
        HTML_TEMPLATE = f.read()

    template = Template(HTML_TEMPLATE)
    html_output = template.render(nodes=nodes_embed, edges=edges_embed, max_degree=50, min_degree=1)
    with open(out_file, "w") as f:
        f.write(html_output)



# (C) Copyright 2024- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import numpy as np
import pytest
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.nodes.attributes.base_attributes import BaseNodeAttribute


class ExtendedBaseNodeAttribute(BaseNodeAttribute):
    """Test implementation of BaseNodeAttribute."""

    def get_raw_values(self, nodes, **_kwargs) -> torch.Tensor:
        return torch.from_numpy(np.array(list(range(nodes.num_nodes))))


@pytest.mark.parametrize("nodes_name", ["invalid_nodes", 4])
def test_base_node_attribute_invalid_nodes_name(graph_with_nodes: HeteroData, nodes_name: str):
    """Test BaseNodeAttribute raises error with invalid nodes name."""
    with pytest.raises(AssertionError):
        ExtendedBaseNodeAttribute().compute(graph_with_nodes, nodes_name)


@pytest.mark.parametrize("norm", ["l3", "invalide"])
def test_base_node_attribute_invalid_norm(graph_with_nodes: HeteroData, norm: str):
    """Test BaseNodeAttribute raises error with invalid nodes name."""
    with pytest.raises(AssertionError):
        ExtendedBaseNodeAttribute(norm=norm).compute(graph_with_nodes, "test_nodes")


@pytest.mark.parametrize("norm", [None, "l1", "l2", "unit-max", "unit-std"])
def test_base_node_attribute_norm(graph_with_nodes: HeteroData, norm: str):
    """Test attribute builder for UniformWeights."""
    node_attr_builder = ExtendedBaseNodeAttribute(norm=norm)
    weights = node_attr_builder.compute(graph_with_nodes, "test_nodes")

    assert weights is not None
    assert isinstance(weights, torch.Tensor)
    assert weights.shape[0] == graph_with_nodes["test_nodes"].x.shape[0]
    assert weights.dtype == node_attr_builder.dtype



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from .builders.cutoff import CutOffEdges
from .builders.cutoff import ReversedCutOffEdges
from .builders.healpix import HEALPixMultiScaleEdges
from .builders.icon import ICONTopologicalDecoderEdges
from .builders.icon import ICONTopologicalEncoderEdges
from .builders.icon import ICONTopologicalProcessorEdges
from .builders.knn import KNNEdges
from .builders.knn import ReversedKNNEdges
from .builders.multi_scale import MultiScaleEdges

__all__ = [
    "KNNEdges",
    "CutOffEdges",
    "HEALPixMultiScaleEdges",
    "MultiScaleEdges",
    "ReversedCutOffEdges",
    "ReversedKNNEdges",
    "ICONTopologicalProcessorEdges",
    "ICONTopologicalEncoderEdges",
    "ICONTopologicalDecoderEdges",
]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import argparse
import logging
from pathlib import Path

from anemoi.graphs.create import GraphCreator
from anemoi.graphs.describe import GraphDescriptor

from . import Command

LOGGER = logging.getLogger(__name__)


class Create(Command):
    """Create a graph."""

    internal = True
    timestamp = True

    def add_arguments(self, command_parser):
        command_parser.add_argument(
            "--overwrite",
            action="store_true",
            help="Overwrite existing files. This will delete the target graph if it already exists.",
        )
        command_parser.add_argument(
            "--description",
            action=argparse.BooleanOptionalAction,
            default=True,
            help="Show the description of the graph.",
        )
        command_parser.add_argument(
            "config",
            type=Path,
            help="Configuration yaml file path defining the recipe to create the graph.",
        )
        command_parser.add_argument("save_path", type=Path, help="Path to store the created graph.")

    def run(self, args):
        if args.save_path.exists() and not args.overwrite:
            LOGGER.warning(
                "Graph already exists at %s. Skipping creation. Use --overwrite to regenerate.",
                args.save_path,
            )
            if args.description:
                GraphDescriptor(args.save_path).describe()
            return

        graph_creator = GraphCreator(config=args.config)
        graph_creator.create(save_path=args.save_path, overwrite=args.overwrite)

        if args.description:
            if args.save_path.exists():
                GraphDescriptor(args.save_path).describe()
            else:
                print("Graph description is not shown if the graph is not saved.")


command = Create



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import time
from abc import ABC
from abc import abstractmethod

import numpy as np
import torch
from hydra.utils import instantiate
from torch_geometric.data import HeteroData

from anemoi.graphs.utils import get_distributed_device
from anemoi.graphs.utils import get_grid_reference_distance
from anemoi.utils.config import DotDict

LOGGER = logging.getLogger(__name__)


class BaseNodeBuilder(ABC):
    """Base class for node builders.

    The node coordinates are stored in the `x` attribute of the nodes and they are stored in radians.

    Attributes
    ----------
    name : str
        name of the nodes, key for the nodes in the HeteroData graph object.
    area_mask_builder : KNNAreaMaskBuilder
        The area of interest mask builder, if any. Defaults to None.
    """

    hidden_attributes: set[str] = set()

    def __init__(self, name: str) -> None:
        self.name = name
        self.area_mask_builder = None
        self.device = get_distributed_device()

    def register_nodes(self, graph: HeteroData) -> HeteroData:
        """Register nodes in the graph.

        Parameters
        ----------
        graph : HeteroData
            The graph to register the nodes.

        Returns
        -------
        HeteroData
            The graph with the registered nodes.
        """
        graph[self.name].x = self.get_coordinates().to(dtype=torch.float32, device=self.device)
        graph[self.name].node_type = type(self).__name__

        if graph[self.name].num_nodes >= 2:
            # At least 2 nodes are needed to compute the grid_reference_distance
            graph[self.name]["_grid_reference_distance"] = get_grid_reference_distance(graph[self.name].x.cpu())
        else:
            LOGGER.warning(f"{self.__class__.__name__} registered {graph[self.name].num_nodes} nodes.")

        return graph

    def register_attributes(self, graph: HeteroData, config: DotDict | None = None) -> HeteroData:
        """Register attributes in the nodes of the graph specified.

        Parameters
        ----------
        graph : HeteroData
            The graph to register the attributes.
        config : DotDict
            The configuration of the attributes.

        Returns
        -------
        HeteroData
            The graph with the registered attributes.
        """
        for hidden_attr in self.hidden_attributes:
            graph[self.name][f"_{hidden_attr}"] = getattr(self, hidden_attr)

        for attr_name, attr_config in config.items():
            graph[self.name][attr_name] = instantiate(attr_config).compute(graph, self.name)

        return graph

    @abstractmethod
    def get_coordinates(self) -> torch.Tensor: ...

    def reshape_coords(
        self, latitudes: np.ndarray | torch.Tensor, longitudes: np.ndarray | torch.Tensor
    ) -> torch.Tensor:
        """Reshape latitude and longitude coordinates.

        Parameters
        ----------
        latitudes : np.ndarray of shape (num_nodes, )
            Latitude coordinates, in degrees.
        longitudes : np.ndarray of shape (num_nodes, )
            Longitude coordinates, in degrees.

        Returns
        -------
        torch.Tensor of shape (num_nodes, 2)
            A 2D tensor with the coordinates, in radians.
        """
        if isinstance(latitudes, np.ndarray):
            latitudes = torch.from_numpy(latitudes)

        if isinstance(longitudes, np.ndarray):
            longitudes = torch.from_numpy(longitudes)

        coords = torch.stack([latitudes, longitudes], axis=-1).reshape((-1, 2))
        return torch.deg2rad(coords)

    def update_graph(self, graph: HeteroData, attrs_config: DotDict | None = None) -> HeteroData:
        """Update the graph with new nodes.

        Parameters
        ----------
        graph : HeteroData
            Input graph.
        attrs_config : DotDict
            The configuration of the attributes.

        Returns
        -------
        HeteroData
            The graph with new nodes included.
        """
        t0 = time.time()
        graph = self.register_nodes(graph)
        t1 = time.time()
        LOGGER.debug("Time to register node coordinates (%s): %.2f s", self.__class__.__name__, t1 - t0)

        t0 = time.time()
        graph = self.register_attributes(graph, attrs_config or {})
        t1 = time.time()
        LOGGER.debug("Time to register node coordinates (%s): %.2f s", self.__class__.__name__, t1 - t0)

        return graph
